## statprof.py
## Copyright (C) 2012 Bryan O'Sullivan <bos@serpentine.com>
## Copyright (C) 2011 Alex Fraser <alex at phatcore dot com>
## Copyright (C) 2004,2005 Andy Wingo <wingo at pobox dot com>
## Copyright (C) 2001 Rob Browning <rlb at defaultvalue dot org>

## This library is free software; you can redistribute it and/or
## modify it under the terms of the GNU Lesser General Public
## License as published by the Free Software Foundation; either
## version 2.1 of the License, or (at your option) any later version.
##
## This library is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
## Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public
## License along with this program; if not, contact:
##
## Free Software Foundation           Voice:  +1-617-542-5942
## 59 Temple Place - Suite 330        Fax:    +1-617-542-2652
## Boston, MA  02111-1307,  USA       gnu@gnu.org

"""
statprof is intended to be a fairly simple statistical profiler for
python. It was ported directly from a statistical profiler for guile,
also named statprof, available from guile-lib [0].

[0] http://wingolog.org/software/guile-lib/statprof/

To start profiling, call statprof.start():
>>> start()

Then run whatever it is that you want to profile, for example:
>>> import test.pystone; test.pystone.pystones()

Then stop the profiling and print out the results:
>>> stop()
>>> display()
  %   cumulative      self
 time    seconds   seconds  name
 26.72      1.40      0.37  pystone.py:79:Proc0
 13.79      0.56      0.19  pystone.py:133:Proc1
 13.79      0.19      0.19  pystone.py:208:Proc8
 10.34      0.16      0.14  pystone.py:229:Func2
  6.90      0.10      0.10  pystone.py:45:__init__
  4.31      0.16      0.06  pystone.py:53:copy
    ...

All of the numerical data is statistically approximate. In the
following column descriptions, and in all of statprof, "time" refers
to execution time (both user and system), not wall clock time.

% time
    The percent of the time spent inside the procedure itself (not
    counting children).

cumulative seconds
    The total number of seconds spent in the procedure, including
    children.

self seconds
    The total number of seconds spent in the procedure itself (not
    counting children).

name
    The name of the procedure.

By default statprof keeps the data collected from previous runs. If you
want to clear the collected data, call reset():
>>> reset()

reset() can also be used to change the sampling frequency from the
default of 1000 Hz. For example, to tell statprof to sample 50 times a
second:
>>> reset(50)

This means that statprof will sample the call stack after every 1/50 of
a second of user + system time spent running on behalf of the python
process. When your process is idle (for example, blocking in a read(),
as is the case at the listener), the clock does not advance. For this
reason statprof is not currently not suitable for profiling io-bound
operations.

The profiler uses the hash of the code object itself to identify the
procedures, so it won't confuse different procedures with the same name.
They will show up as two different rows in the output.

Right now the profiler is quite simplistic.  I cannot provide
call-graphs or other higher level information.  What you see in the
table is pretty much all there is. Patches are welcome :-)


Threading
---------

Because signals only get delivered to the main thread in Python,
statprof only profiles the main thread. However because the time
reporting function uses per-process timers, the results can be
significantly off if other threads' work patterns are not similar to the
main thread's work patterns.
"""
# no-check-code
from __future__ import annotations

import collections
import contextlib
import getopt
import inspect
import json
import os
import signal
import sys
import threading
import time

from . import (
    encoding,
    pycompat,
)

defaultdict = collections.defaultdict
contextmanager = contextlib.contextmanager

__all__ = ['start', 'stop', 'reset', 'display', 'profile']

skips = {
    "util.py:check",
    "extensions.py:closure",
    "color.py:colorcmd",
    "dispatch.py:checkargs",
    "dispatch.py:<lambda>",
    "dispatch.py:_runcatch",
    "dispatch.py:_dispatch",
    "dispatch.py:_runcommand",
    "pager.py:pagecmd",
    "dispatch.py:run",
    "dispatch.py:dispatch",
    "dispatch.py:runcommand",
    "hg.py:<module>",
    "evolve.py:warnobserrors",
}

###########################################################################
## Utils


def clock():
    times = os.times()
    return (times[0] + times[1], times[4])


###########################################################################
## Collection data structures


class ProfileState:
    samples: list[Sample]

    def __init__(self, frequency=None):
        self.reset(frequency)
        self.track = b'cpu'

    def reset(self, frequency=None):
        # total so far
        self.accumulated_time = (0.0, 0.0)
        # start_time when timer is active
        self.last_start_time = None
        # a float
        if frequency:
            self.sample_interval = 1.0 / frequency
        elif not hasattr(self, 'sample_interval'):
            # default to 1000 Hz
            self.sample_interval = 1.0 / 1000.0
        else:
            # leave the frequency as it was
            pass
        self.remaining_prof_time = None
        # for user start/stop nesting
        self.profile_level = 0

        self.samples = []

    def accumulate_time(self, stop_time):
        increment = (
            stop_time[0] - self.last_start_time[0],
            stop_time[1] - self.last_start_time[1],
        )
        self.accumulated_time = (
            self.accumulated_time[0] + increment[0],
            self.accumulated_time[1] + increment[1],
        )

    def seconds_per_sample(self):
        return self.accumulated_time[self.timeidx] / len(self.samples)

    @property
    def timeidx(self):
        if self.track == b'real':
            return 1
        return 0


state = ProfileState()


class CodeSite:
    cache = {}

    __slots__ = ('path', 'lineno', 'function', 'source')

    def __init__(self, path, lineno, function):
        assert isinstance(path, bytes)
        self.path = path
        self.lineno = lineno
        assert isinstance(function, bytes)
        self.function = function
        self.source = None

    def __eq__(self, other):
        try:
            return self.lineno == other.lineno and self.path == other.path
        except:
            return False

    def __hash__(self):
        return hash((self.lineno, self.path))

    @classmethod
    def get(cls, path, lineno, function):
        k = (path, lineno)
        try:
            return cls.cache[k]
        except KeyError:
            v = cls(path, lineno, function)
            cls.cache[k] = v
            return v

    def getsource(self, length):
        if self.source is None:
            try:
                lineno = self.lineno - 1  # lineno can be None
                with open(self.path, 'rb') as fp:
                    for i, line in enumerate(fp):
                        if i == lineno:
                            self.source = line.strip()
                            break
            except:
                pass
            if self.source is None:
                self.source = b''

        source = self.source
        if len(source) > length:
            source = source[: (length - 3)] + b"..."
        return source

    def filename(self):
        return os.path.basename(self.path)

    def skipname(self):
        return '%s:%s' % (self.filename(), self.function)


class Sample:
    __slots__ = ('stack', 'time')

    def __init__(self, stack, time):
        self.stack = stack
        self.time = time

    @classmethod
    def from_frame(cls, frame, time):
        stack = []

        while frame:
            stack.append(
                CodeSite.get(
                    pycompat.sysbytes(frame.f_code.co_filename),
                    frame.f_lineno,
                    pycompat.sysbytes(frame.f_code.co_name),
                )
            )
            frame = frame.f_back

        return Sample(stack, time)


###########################################################################
## SIGPROF handler


def profile_signal_handler(signum, frame):
    if state.profile_level > 0:
        now = clock()
        state.accumulate_time(now)

        timestamp = state.accumulated_time[state.timeidx]
        state.samples.append(Sample.from_frame(frame, timestamp))

        signal.setitimer(signal.ITIMER_PROF, state.sample_interval, 0.0)
        state.last_start_time = now


stopthread = threading.Event()


def samplerthread(tid):
    while not stopthread.is_set():
        now = clock()
        state.accumulate_time(now)

        frame = sys._current_frames()[tid]

        timestamp = state.accumulated_time[state.timeidx]
        state.samples.append(Sample.from_frame(frame, timestamp))

        state.last_start_time = now
        time.sleep(state.sample_interval)

    stopthread.clear()


###########################################################################
## Profiling API


def is_active():
    return state.profile_level > 0


lastmechanism = None


def start(mechanism=b'thread', track=b'cpu'):
    '''Install the profiling signal handler, and start profiling.'''
    state.track = track  # note: nesting different mode won't work
    state.profile_level += 1
    if state.profile_level == 1:
        state.last_start_time = clock()
        rpt = state.remaining_prof_time
        state.remaining_prof_time = None

        global lastmechanism
        lastmechanism = mechanism

        if mechanism == b'signal':
            signal.signal(signal.SIGPROF, profile_signal_handler)
            signal.setitimer(
                signal.ITIMER_PROF, rpt or state.sample_interval, 0.0
            )
        elif mechanism == b'thread':
            frame = inspect.currentframe()
            tid = [k for k, f in sys._current_frames().items() if f == frame][0]
            state.thread = threading.Thread(
                target=samplerthread, args=(tid,), name="samplerthread"
            )
            state.thread.start()


def stop():
    '''Stop profiling, and uninstall the profiling signal handler.'''
    state.profile_level -= 1
    if state.profile_level == 0:
        if lastmechanism == b'signal':
            rpt = signal.setitimer(signal.ITIMER_PROF, 0.0, 0.0)
            signal.signal(signal.SIGPROF, signal.SIG_IGN)
            state.remaining_prof_time = rpt[0]
        elif lastmechanism == b'thread':
            stopthread.set()
            state.thread.join()

        state.accumulate_time(clock())
        state.last_start_time = None
        statprofpath = encoding.environ.get(b'STATPROF_DEST')
        if statprofpath:
            save_data(statprofpath)

    return state


def save_data(path):
    with open(path, 'w+b') as file:
        file.write(b"%f %f\n" % state.accumulated_time)
        for sample in state.samples:
            time = sample.time
            stack = sample.stack
            sites = [
                b'\1'.join([s.path, b'%d' % s.lineno or -1, s.function])
                for s in stack
            ]
            file.write(b"%d\0%s\n" % (time, b'\0'.join(sites)))


def load_data(path):
    with open(path, 'rb') as fp:
        lines = fp.read().splitlines()

    state.accumulated_time = [float(value) for value in lines[0].split()]
    state.samples = []
    for line in lines[1:]:
        parts = line.split(b'\0')
        time = float(parts[0])
        rawsites = parts[1:]
        sites = []
        for rawsite in rawsites:
            siteparts = rawsite.split(b'\1')
            sites.append(
                CodeSite.get(siteparts[0], int(siteparts[1]), siteparts[2])
            )

        state.samples.append(Sample(sites, time))


def reset(frequency=None):
    """Clear out the state of the profiler.  Do not call while the
    profiler is running.

    The optional frequency argument specifies the number of samples to
    collect per second."""
    assert state.profile_level == 0, b"Can't reset() while statprof is running"
    CodeSite.cache.clear()
    state.reset(frequency)


@contextmanager
def profile():
    start()
    try:
        yield
    finally:
        stop()
        display()


###########################################################################
## Reporting API


class SiteStats:
    def __init__(self, site):
        self.site = site
        self.selfcount = 0
        self.totalcount = 0

    def addself(self):
        self.selfcount += 1

    def addtotal(self):
        self.totalcount += 1

    def selfpercent(self):
        return self.selfcount / len(state.samples) * 100

    def totalpercent(self):
        return self.totalcount / len(state.samples) * 100

    def selfseconds(self):
        return self.selfcount * state.seconds_per_sample()

    def totalseconds(self):
        return self.totalcount * state.seconds_per_sample()

    @classmethod
    def buildstats(cls, samples):
        stats = {}

        for sample in samples:
            for i, site in enumerate(sample.stack):
                sitestat = stats.get(site)
                if not sitestat:
                    sitestat = SiteStats(site)
                    stats[site] = sitestat

                sitestat.addtotal()

                if i == 0:
                    sitestat.addself()

        return [s for s in stats.values()]


class DisplayFormats:
    ByLine = 0
    ByMethod = 1
    AboutMethod = 2
    Hotpath = 3
    FlameGraph = 4
    Json = 5
    Chrome = 6


def display(fp=None, format=3, data=None, **kwargs):
    '''Print statistics, either to stdout or the given file object.'''
    if data is None:
        data = state

    if fp is None:
        from .utils import procutil

        fp = procutil.stdout
    if len(data.samples) == 0:
        fp.write(b'No samples recorded.\n')
        return

    if format == DisplayFormats.ByLine:
        display_by_line(data, fp)
    elif format == DisplayFormats.ByMethod:
        display_by_method(data, fp)
    elif format == DisplayFormats.AboutMethod:
        display_about_method(data, fp, **kwargs)
    elif format == DisplayFormats.Hotpath:
        display_hotpath(data, fp, **kwargs)
    elif format == DisplayFormats.FlameGraph:
        write_to_flame(data, fp, **kwargs)
    elif format == DisplayFormats.Json:
        write_to_json(data, fp)
    elif format == DisplayFormats.Chrome:
        write_to_chrome(data, fp, **kwargs)
    else:
        raise Exception("Invalid display format")

    if format not in (DisplayFormats.Json, DisplayFormats.Chrome):
        fp.write(b'---\n')
        fp.write(b'Sample count: %d\n' % len(data.samples))
        fp.write(b'Total time: %f seconds (%f wall)\n' % data.accumulated_time)


def display_by_line(data, fp):
    """Print the profiler data with each sample line represented
    as one row in a table.  Sorted by self-time per line."""
    stats = SiteStats.buildstats(data.samples)
    stats.sort(reverse=True, key=lambda x: x.selfseconds())

    fp.write(
        b'%5.5s %10.10s   %7.7s  %-8.8s\n'
        % (b'%  ', b'cumulative', b'self', b'')
    )
    fp.write(
        b'%5.5s  %9.9s  %8.8s  %-8.8s\n'
        % (b"time", b"seconds", b"seconds", b"name")
    )

    for stat in stats:
        site = stat.site
        sitelabel = b'%s:%d:%s' % (
            site.filename(),
            site.lineno or -1,
            site.function,
        )
        fp.write(
            b'%6.2f %9.2f %9.2f  %s\n'
            % (
                stat.selfpercent(),
                stat.totalseconds(),
                stat.selfseconds(),
                sitelabel,
            )
        )


def display_by_method(data, fp):
    """Print the profiler data with each sample function represented
    as one row in a table.  Important lines within that function are
    output as nested rows.  Sorted by self-time per line."""
    fp.write(
        b'%5.5s %10.10s   %7.7s  %-8.8s\n'
        % (b'%  ', b'cumulative', b'self', b'')
    )
    fp.write(
        b'%5.5s  %9.9s  %8.8s  %-8.8s\n'
        % (b"time", b"seconds", b"seconds", b"name")
    )

    stats = SiteStats.buildstats(data.samples)

    grouped = defaultdict(list)
    for stat in stats:
        grouped[stat.site.filename() + b":" + stat.site.function].append(stat)

    # compute sums for each function
    functiondata = []
    for fname, sitestats in grouped.items():
        total_cum_sec = 0
        total_self_sec = 0
        total_percent = 0
        for stat in sitestats:
            total_cum_sec += stat.totalseconds()
            total_self_sec += stat.selfseconds()
            total_percent += stat.selfpercent()

        functiondata.append(
            (fname, total_cum_sec, total_self_sec, total_percent, sitestats)
        )

    # sort by total self sec
    functiondata.sort(reverse=True, key=lambda x: x[2])

    for function in functiondata:
        if function[3] < 0.05:
            continue
        fp.write(
            b'%6.2f %9.2f %9.2f  %s\n'
            % (
                function[3],  # total percent
                function[1],  # total cum sec
                function[2],  # total self sec
                function[0],
            )
        )  # file:function

        function[4].sort(reverse=True, key=lambda i: i.selfseconds())
        for stat in function[4]:
            # only show line numbers for significant locations (>1% time spent)
            if stat.selfpercent() > 1:
                source = stat.site.getsource(25)
                if not isinstance(source, bytes):
                    source = pycompat.bytestr(source)

                stattuple = (
                    stat.selfpercent(),
                    stat.selfseconds(),
                    stat.site.lineno or -1,
                    source,
                )

                fp.write(b'%33.0f%% %6.2f   line %d: %s\n' % stattuple)


def display_about_method(data, fp, function=None, **kwargs):
    if function is None:
        raise Exception("Invalid function")

    filename = None
    if b':' in function:
        filename, function = function.split(b':')

    relevant_samples = 0
    parents = {}
    children = {}

    for sample in data.samples:
        for i, site in enumerate(sample.stack):
            if site.function == function and (
                not filename or site.filename() == filename
            ):
                relevant_samples += 1
                if i != len(sample.stack) - 1:
                    parent = sample.stack[i + 1]
                    if parent in parents:
                        parents[parent] = parents[parent] + 1
                    else:
                        parents[parent] = 1

                if site in children:
                    children[site] = children[site] + 1
                else:
                    children[site] = 1

    parents = [(parent, count) for parent, count in parents.items()]
    parents.sort(reverse=True, key=lambda x: x[1])
    for parent, count in parents:
        fp.write(
            b'%6.2f%%   %s:%s   line %s: %s\n'
            % (
                count / relevant_samples * 100,
                pycompat.fsencode(parent.filename()),
                pycompat.sysbytes(parent.function),
                parent.lineno or -1,
                pycompat.sysbytes(parent.getsource(50)),
            )
        )

    stats = SiteStats.buildstats(data.samples)
    stats = [
        s
        for s in stats
        if s.site.function == function
        and (not filename or s.site.filename() == filename)
    ]

    total_cum_sec = 0
    total_self_sec = 0
    total_self_percent = 0
    total_cum_percent = 0
    for stat in stats:
        total_cum_sec += stat.totalseconds()
        total_self_sec += stat.selfseconds()
        total_self_percent += stat.selfpercent()
        total_cum_percent += stat.totalpercent()

    fp.write(
        b'\n    %s:%s    Total: %0.2fs (%0.2f%%)    Self: %0.2fs (%0.2f%%)\n\n'
        % (
            pycompat.sysbytes(filename or b'___'),
            pycompat.sysbytes(function),
            total_cum_sec,
            total_cum_percent,
            total_self_sec,
            total_self_percent,
        )
    )

    children = [(child, count) for child, count in children.items()]
    children.sort(reverse=True, key=lambda x: x[1])
    for child, count in children:
        fp.write(
            b'        %6.2f%%   line %s: %s\n'
            % (
                count / relevant_samples * 100,
                child.lineno or -1,
                pycompat.sysbytes(child.getsource(50)),
            )
        )


def display_hotpath(data, fp, limit=0.05, **kwargs):
    class HotNode:
        def __init__(self, site):
            self.site = site
            self.count = 0
            self.children = {}

        def add(self, stack, time):
            self.count += time
            site = stack[0]
            child = self.children.get(site)
            if not child:
                child = HotNode(site)
                self.children[site] = child

            if len(stack) > 1:
                i = 1
                # Skip boiler plate parts of the stack
                while i < len(stack) and stack[i].skipname() in skips:
                    i += 1
                if i < len(stack):
                    child.add(stack[i:], time)
            else:
                # Normally this is done by the .add() calls
                child.count += time

    root = HotNode(None)
    lasttime = data.samples[0].time
    for sample in data.samples:
        root.add(sample.stack[::-1], sample.time - lasttime)
        lasttime = sample.time
    showtime = kwargs.get('showtime', True)

    def _write(node, depth, multiple_siblings):
        site = node.site
        visiblechildren = [
            c for c in node.children.values() if c.count >= (limit * root.count)
        ]
        if site:
            indent = depth * 2 - 1
            filename = (site.filename() + b':').ljust(15)
            function = site.function

            # lots of string formatting
            listpattern = (
                b''.ljust(indent)
                + (b'\\' if multiple_siblings else b'|')
                + b' %4.1f%%'
                + (b' %5.2fs' % node.count if showtime else b'')
                + b'  %s %s'
            )
            liststring = listpattern % (
                node.count / root.count * 100,
                filename,
                function,
            )
            # 4 to account for the word 'line'
            spacing_len = max(4, 55 - len(liststring))
            prefix = b''
            if spacing_len == 4:
                prefix = b', '

            codepattern = b'%s%s %d: %s%s'
            codestring = codepattern % (
                prefix,
                b'line'.rjust(spacing_len),
                site.lineno if site.lineno is not None else -1,
                b''.ljust(max(0, 4 - len(str(site.lineno)))),
                site.getsource(30),
            )

            finalstring = liststring + codestring
            childrensamples = sum([c.count for c in node.children.values()])
            # Make frames that performed more than 10% of the operation red
            if node.count - childrensamples > (0.1 * root.count):
                finalstring = b'\033[91m' + finalstring + b'\033[0m'
            # Make frames that didn't actually perform work dark grey
            elif node.count - childrensamples == 0:
                finalstring = b'\033[90m' + finalstring + b'\033[0m'
            fp.write(finalstring + b'\n')

        newdepth = depth
        if len(visiblechildren) > 1 or multiple_siblings:
            newdepth += 1

        visiblechildren.sort(reverse=True, key=lambda x: x.count)
        for child in visiblechildren:
            _write(child, newdepth, len(visiblechildren) > 1)

    if root.count > 0:
        _write(root, 0, False)


def write_to_flame(data, fp, scriptpath=None, outputfile=None, **kwargs):
    if scriptpath is None:
        scriptpath = encoding.environ[b'HOME'] + b'/flamegraph.pl'
    if not os.path.exists(scriptpath):
        fp.write(b'error: missing %s\n' % scriptpath)
        fp.write(b'get it here: https://github.com/brendangregg/FlameGraph\n')
        return

    lines = {}
    for sample in data.samples:
        sites = [s.function for s in sample.stack]
        sites.reverse()
        line = b';'.join(sites)
        if line in lines:
            lines[line] = lines[line] + 1
        else:
            lines[line] = 1

    fd, path = pycompat.mkstemp()

    with open(path, "w+b") as file:
        for line, count in lines.items():
            file.write(b"%s %d\n" % (line, count))

    if outputfile is None:
        outputfile = b'~/flamegraph.svg'

    os.system(b"perl ~/flamegraph.pl %s > %s" % (path, outputfile))
    fp.write(b'Written to %s\n' % outputfile)


_pathcache = {}


def simplifypath(path):
    """Attempt to make the path to a Python module easier to read by
    removing whatever part of the Python search path it was found
    on."""

    if path in _pathcache:
        return _pathcache[path]
    hgpath = encoding.__file__.rsplit(os.sep, 2)[0]
    for p in [hgpath] + sys.path:
        prefix = p + os.sep
        if path.startswith(prefix):
            path = path[len(prefix) :]
            break
    _pathcache[path] = path
    return path


def write_to_json(data, fp):
    samples = []

    for sample in data.samples:
        stack = []

        for frame in sample.stack:
            stack.append(
                (
                    pycompat.sysstr(frame.path),
                    frame.lineno or -1,
                    pycompat.sysstr(frame.function),
                )
            )

        samples.append((sample.time, stack))

    data = json.dumps(samples)
    if not isinstance(data, bytes):
        data = data.encode('utf-8')

    fp.write(data)


def write_to_chrome(data, fp, minthreshold=0.005, maxthreshold=0.999):
    samples = []
    laststack = collections.deque()
    lastseen = collections.deque()

    # The Chrome tracing format allows us to use a compact stack
    # representation to save space. It's fiddly but worth it.
    # We maintain a bijection between stack and ID.
    stack2id = {}
    id2stack = []  # will eventually be rendered

    def stackid(stack):
        if not stack:
            return
        if stack in stack2id:
            return stack2id[stack]
        parent = stackid(stack[1:])
        myid = len(stack2id)
        stack2id[stack] = myid
        id2stack.append(dict(category=stack[0][0], name='%s %s' % stack[0]))
        if parent is not None:
            id2stack[-1].update(parent=parent)
        return myid

    # The sampling profiler can sample multiple times without
    # advancing the clock, potentially causing the Chrome trace viewer
    # to render single-pixel columns that we cannot zoom in on.  We
    # work around this by pretending that zero-duration samples are a
    # millisecond in length.

    clamp = 0.001

    # We provide knobs that by default attempt to filter out stack
    # frames that are too noisy:
    #
    # * A few take almost all execution time. These are usually boring
    #   setup functions, giving a stack that is deep but uninformative.
    #
    # * Numerous samples take almost no time, but introduce lots of
    #   noisy, oft-deep "spines" into a rendered profile.

    blacklist = set()
    totaltime = data.samples[-1].time - data.samples[0].time
    minthreshold = totaltime * minthreshold
    maxthreshold = max(totaltime * maxthreshold, clamp)

    def poplast():
        oldsid = stackid(tuple(laststack))
        oldcat, oldfunc = laststack.popleft()
        oldtime, oldidx = lastseen.popleft()
        duration = sample.time - oldtime
        if minthreshold <= duration <= maxthreshold:
            # ensure no zero-duration events
            sampletime = max(oldtime + clamp, sample.time)
            samples.append(
                dict(
                    ph='E',
                    name=oldfunc,
                    cat=oldcat,
                    sf=oldsid,
                    ts=sampletime * 1e6,
                    pid=0,
                )
            )
        else:
            blacklist.add(oldidx)

    # Much fiddling to synthesize correctly(ish) nested begin/end
    # events given only stack snapshots.

    for sample in data.samples:
        stack = tuple(
            (
                '%s:%d'
                % (
                    simplifypath(pycompat.sysstr(frame.path)),
                    frame.lineno or -1,
                ),
                pycompat.sysstr(frame.function),
            )
            for frame in sample.stack
        )
        qstack = collections.deque(stack)
        if laststack == qstack:
            continue
        while laststack and qstack and laststack[-1] == qstack[-1]:
            laststack.pop()
            qstack.pop()
        while laststack:
            poplast()
        for f in reversed(qstack):
            lastseen.appendleft((sample.time, len(samples)))
            laststack.appendleft(f)
            path, name = f
            sid = stackid(tuple(laststack))
            samples.append(
                dict(
                    ph='B',
                    name=name,
                    cat=path,
                    ts=sample.time * 1e6,
                    sf=sid,
                    pid=0,
                )
            )
        laststack = collections.deque(stack)
    while laststack:
        poplast()
    events = [
        sample for idx, sample in enumerate(samples) if idx not in blacklist
    ]
    frames = collections.OrderedDict(
        (str(k), v) for (k, v) in enumerate(id2stack)
    )
    data = json.dumps(dict(traceEvents=events, stackFrames=frames), indent=1)
    if not isinstance(data, bytes):
        data = data.encode('utf-8')
    fp.write(data)
    fp.write(b'\n')


def printusage():
    print(
        r"""
The statprof command line allows you to inspect the last profile's results in
the following forms:

usage:
    hotpath [-l --limit percent]
        Shows a graph of calls with the percent of time each takes.
        Red calls take over 10%% of the total time themselves.
    lines
        Shows the actual sampled lines.
    functions
        Shows the samples grouped by function.
    function [filename:]functionname
        Shows the callers and callees of a particular function.
    flame [-s --script-path] [-o --output-file path]
        Writes out a flamegraph to output-file (defaults to ~/flamegraph.svg)
        Requires that ~/flamegraph.pl exist.
        (Specify alternate script path with --script-path.)"""
    )


def main(argv=None):
    if argv is None:
        argv = sys.argv

    if len(argv) == 1:
        printusage()
        return 0

    displayargs = {}

    optstart = 2
    displayargs[b'function'] = None
    if argv[1] == 'hotpath':
        displayargs[b'format'] = DisplayFormats.Hotpath
    elif argv[1] == 'lines':
        displayargs[b'format'] = DisplayFormats.ByLine
    elif argv[1] == 'functions':
        displayargs[b'format'] = DisplayFormats.ByMethod
    elif argv[1] == 'function':
        displayargs[b'format'] = DisplayFormats.AboutMethod
        displayargs[b'function'] = argv[2]
        optstart = 3
    elif argv[1] == 'flame':
        displayargs[b'format'] = DisplayFormats.FlameGraph
    else:
        printusage()
        return 0

    # process options
    try:
        opts, args = pycompat.getoptb(
            pycompat.sysargv[optstart:],
            b"hl:f:o:p:",
            [b"help", b"limit=", b"file=", b"output-file=", b"script-path="],
        )
    except getopt.error as msg:
        print(msg)
        printusage()
        return 2

    displayargs[b'limit'] = 0.05
    path = None
    for o, value in opts:
        if o in ("-l", "--limit"):
            displayargs[b'limit'] = float(value)
        elif o in ("-f", "--file"):
            path = value
        elif o in ("-o", "--output-file"):
            displayargs[b'outputfile'] = value
        elif o in ("-p", "--script-path"):
            displayargs[b'scriptpath'] = value
        elif o in ("-h", "help"):
            printusage()
            return 0
        else:
            assert False, "unhandled option %s" % o

    if not path:
        print('must specify --file to load')
        return 1

    load_data(path=path)

    display(**pycompat.strkwargs(displayargs))

    return 0


if __name__ == "__main__":
    sys.exit(main())



# Copyright 2016-present Facebook. All Rights Reserved.
#
# revmap: trivial hg hash - linelog rev bidirectional map
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.

from __future__ import annotations

import bisect
import io
import os
import struct

from mercurial.node import hex
from mercurial import (
    error as hgerror,
)
from . import error

# the revmap file format is straightforward:
#
#    8 bytes: header
#    1 byte : flag for linelog revision 1
#    ? bytes: (optional) '\0'-terminated path string
#             only exists if (flag & renameflag) != 0
#   20 bytes: hg hash for linelog revision 1
#    1 byte : flag for linelog revision 2
#    ? bytes: (optional) '\0'-terminated path string
#   20 bytes: hg hash for linelog revision 2
#   ....
#
# the implementation is kinda stupid: __init__ loads the whole revmap.
# no laziness. benchmark shows loading 10000 revisions is about 0.015
# seconds, which looks enough for our use-case. if this implementation
# becomes a bottleneck, we can change it to lazily read the file
# from the end.

# whether the changeset is in the side branch. i.e. not in the linear main
# branch but only got referenced by lines in merge changesets.
sidebranchflag = 1

# whether the changeset changes the file path (ie. is a rename)
renameflag = 2

# len(mercurial.node.nullid)
_hshlen = 20


class revmap:
    """trivial hg bin hash - linelog rev bidirectional map

    also stores a flag (uint8) for each revision, and track renames.
    """

    HEADER = b'REVMAP1\0'

    def __init__(self, path=None):
        """create or load the revmap, optionally associate to a file

        if path is None, the revmap is entirely in-memory. the caller is
        responsible for locking. concurrent writes to a same file is unsafe.
        the caller needs to make sure one file is associated to at most one
        revmap object at a time."""
        self.path = path
        self._rev2hsh = [None]
        self._rev2flag = [None]
        self._hsh2rev = {}
        # since rename does not happen frequently, do not store path for every
        # revision. self._renamerevs can be used for bisecting.
        self._renamerevs = [0]
        self._renamepaths = [b'']
        self._lastmaxrev = -1
        if path:
            if os.path.exists(path):
                self._load()
            else:
                # write the header so "append" can do incremental updates
                self.flush()

    def copyfrom(self, rhs):
        """copy the map data from another revmap. do not affect self.path"""
        self._rev2hsh = rhs._rev2hsh[:]
        self._rev2flag = rhs._rev2flag[:]
        self._hsh2rev = rhs._hsh2rev.copy()
        self._renamerevs = rhs._renamerevs[:]
        self._renamepaths = rhs._renamepaths[:]
        self._lastmaxrev = -1

    @property
    def maxrev(self):
        """return max linelog revision number"""
        return len(self._rev2hsh) - 1

    def append(self, hsh, sidebranch=False, path=None, flush=False):
        """add a binary hg hash and return the mapped linelog revision.
        if flush is True, incrementally update the file.
        """
        if hsh in self._hsh2rev:
            raise error.CorruptedFileError(
                b'%r is in revmap already' % hex(hsh)
            )
        if len(hsh) != _hshlen:
            raise hgerror.ProgrammingError(
                b'hsh must be %d-char long' % _hshlen
            )
        idx = len(self._rev2hsh)
        flag = 0
        if sidebranch:
            flag |= sidebranchflag
        if path is not None and path != self._renamepaths[-1]:
            flag |= renameflag
            self._renamerevs.append(idx)
            self._renamepaths.append(path)
        self._rev2hsh.append(hsh)
        self._rev2flag.append(flag)
        self._hsh2rev[hsh] = idx
        if flush:
            self.flush()
        return idx

    def rev2hsh(self, rev):
        """convert linelog revision to hg hash. return None if not found."""
        if rev > self.maxrev or rev < 0:
            return None
        return self._rev2hsh[rev]

    def rev2flag(self, rev):
        """get the flag (uint8) for a given linelog revision.
        return None if revision does not exist.
        """
        if rev > self.maxrev or rev < 0:
            return None
        return self._rev2flag[rev]

    def rev2path(self, rev):
        """get the path for a given linelog revision.
        return None if revision does not exist.
        """
        if rev > self.maxrev or rev < 0:
            return None
        idx = bisect.bisect_right(self._renamerevs, rev) - 1
        return self._renamepaths[idx]

    def hsh2rev(self, hsh):
        """convert hg hash to linelog revision. return None if not found."""
        return self._hsh2rev.get(hsh)

    def clear(self, flush=False):
        """make the map empty. if flush is True, write to disk"""
        # rev 0 is reserved, real rev starts from 1
        self._rev2hsh = [None]
        self._rev2flag = [None]
        self._hsh2rev = {}
        self._rev2path = [b'']
        self._lastmaxrev = -1
        if flush:
            self.flush()

    def flush(self):
        """write the state down to the file"""
        if not self.path:
            return
        if self._lastmaxrev == -1:  # write the entire file
            with open(self.path, 'wb') as f:
                f.write(self.HEADER)
                for i in range(1, len(self._rev2hsh)):
                    self._writerev(i, f)
        else:  # append incrementally
            with open(self.path, 'ab') as f:
                for i in range(self._lastmaxrev + 1, len(self._rev2hsh)):
                    self._writerev(i, f)
        self._lastmaxrev = self.maxrev

    def _load(self):
        """load state from file"""
        if not self.path:
            return
        # use local variables in a loop. CPython uses LOAD_FAST for them,
        # which is faster than both LOAD_CONST and LOAD_GLOBAL.
        flaglen = 1
        hshlen = _hshlen
        with open(self.path, 'rb') as f:
            if f.read(len(self.HEADER)) != self.HEADER:
                raise error.CorruptedFileError()
            self.clear(flush=False)
            while True:
                buf = f.read(flaglen)
                if not buf:
                    break
                flag = ord(buf)
                rev = len(self._rev2hsh)
                if flag & renameflag:
                    path = self._readcstr(f)
                    self._renamerevs.append(rev)
                    self._renamepaths.append(path)
                hsh = f.read(hshlen)
                if len(hsh) != hshlen:
                    raise error.CorruptedFileError()
                self._hsh2rev[hsh] = rev
                self._rev2flag.append(flag)
                self._rev2hsh.append(hsh)
        self._lastmaxrev = self.maxrev

    def _writerev(self, rev, f):
        """append a revision data to file"""
        flag = self._rev2flag[rev]
        hsh = self._rev2hsh[rev]
        f.write(struct.pack(b'B', flag))
        if flag & renameflag:
            path = self.rev2path(rev)
            if path is None:
                raise error.CorruptedFileError(b'cannot find path for %s' % rev)
            f.write(path + b'\0')
        f.write(hsh)

    @staticmethod
    def _readcstr(f):
        """read a C-language-like '\0'-terminated string"""
        buf = b''
        while True:
            ch = f.read(1)
            if not ch:  # unexpected eof
                raise error.CorruptedFileError()
            if ch == b'\0':
                break
            buf += ch
        return buf

    def __contains__(self, f):
        """(fctx or (node, path)) -> bool.
        test if (node, path) is in the map, and is not in a side branch.
        f can be either a tuple of (node, path), or a fctx.
        """
        if isinstance(f, tuple):  # f: (node, path)
            hsh, path = f
        else:  # f: fctx
            hsh, path = f.node(), f.path()
        rev = self.hsh2rev(hsh)
        if rev is None:
            return False
        if path is not None and path != self.rev2path(rev):
            return False
        return (self.rev2flag(rev) & sidebranchflag) == 0


def getlastnode(path):
    """return the last hash in a revmap, without loading its full content.
    this is equivalent to `m = revmap(path); m.rev2hsh(m.maxrev)`, but faster.
    """
    hsh = None
    try:
        with open(path, 'rb') as f:
            f.seek(-_hshlen, io.SEEK_END)
            if f.tell() > len(revmap.HEADER):
                hsh = f.read(_hshlen)
    except OSError:
        pass
    return hsh



# bundle2.py - generic container format to transmit arbitrary data.
#
# Copyright 2013 Facebook, Inc.
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
"""Handling of the new bundle2 format

The goal of bundle2 is to act as an atomically packet to transmit a set of
payloads in an application agnostic way. It consist in a sequence of "parts"
that will be handed to and processed by the application layer.


General format architecture
===========================

The format is architectured as follow

 - magic string
 - stream level parameters
 - payload parts (any number)
 - end of stream marker.

the Binary format
============================

All numbers are unsigned and big-endian.

stream level parameters
------------------------

Binary format is as follow

:params size: int32

  The total number of Bytes used by the parameters

:params value: arbitrary number of Bytes

  A blob of `params size` containing the serialized version of all stream level
  parameters.

  The blob contains a space separated list of parameters. Parameters with value
  are stored in the form `<name>=<value>`. Both name and value are urlquoted.

  Empty name are obviously forbidden.

  Name MUST start with a letter. If this first letter is lower case, the
  parameter is advisory and can be safely ignored. However when the first
  letter is capital, the parameter is mandatory and the bundling process MUST
  stop if he is not able to proceed it.

  Stream parameters use a simple textual format for two main reasons:

  - Stream level parameters should remain simple and we want to discourage any
    crazy usage.
  - Textual data allow easy human inspection of a bundle2 header in case of
    troubles.

  Any Applicative level options MUST go into a bundle2 part instead.

Payload part
------------------------

Binary format is as follow

:header size: int32

  The total number of Bytes used by the part header. When the header is empty
  (size = 0) this is interpreted as the end of stream marker.

:header:

    The header defines how to interpret the part. It contains two piece of
    data: the part type, and the part parameters.

    The part type is used to route an application level handler, that can
    interpret payload.

    Part parameters are passed to the application level handler.  They are
    meant to convey information that will help the application level object to
    interpret the part payload.

    The binary format of the header is has follow

    :typesize: (one byte)

    :parttype: alphanumerical part name (restricted to [a-zA-Z0-9_:-]*)

    :partid: A 32bits integer (unique in the bundle) that can be used to refer
             to this part.

    :parameters:

        Part's parameter may have arbitrary content, the binary structure is::

            <mandatory-count><advisory-count><param-sizes><param-data>

        :mandatory-count: 1 byte, number of mandatory parameters

        :advisory-count:  1 byte, number of advisory parameters

        :param-sizes:

            N couple of bytes, where N is the total number of parameters. Each
            couple contains (<size-of-key>, <size-of-value) for one parameter.

        :param-data:

            A blob of bytes from which each parameter key and value can be
            retrieved using the list of size couples stored in the previous
            field.

            Mandatory parameters comes first, then the advisory ones.

            Each parameter's key MUST be unique within the part.

:payload:

    payload is a series of `<chunksize><chunkdata>`.

    `chunksize` is an int32, `chunkdata` are plain bytes (as much as
    `chunksize` says)` The payload part is concluded by a zero size chunk.

    The current implementation always produces either zero or one chunk.
    This is an implementation limitation that will ultimately be lifted.

    `chunksize` can be negative to trigger special case processing. No such
    processing is in place yet.

Bundle processing
============================

Each part is processed in order using a "part handler". Handler are registered
for a certain part type.

The matching of a part to its handler is case insensitive. The case of the
part type is used to know if a part is mandatory or advisory. If the Part type
contains any uppercase char it is considered mandatory. When no handler is
known for a Mandatory part, the process is aborted and an exception is raised.
If the part is advisory and no handler is known, the part is ignored. When the
process is aborted, the full bundle is still read from the stream to keep the
channel usable. But none of the part read from an abort are processed. In the
future, dropping the stream may become an option for channel we do not care to
preserve.
"""

from __future__ import annotations

import collections
import errno
import os
import re
import string
import struct
import sys
import typing

from .i18n import _
from .interfaces.types import (
    Capabilities,
    MatcherT,
    RepoT,
    UnbundleOpT,
)
from . import (
    changegroup,
    encoding,
    error,
    obsolete,
    phases,
    policy,
    pycompat,
    scmutil,
    streamclone,
    tables,
    tags,
    util,
    vfs as vfsmod,
)
from .exchanges import (
    bundle_caps,
)
from .utils import (
    stringutil,
    urlutil,
)
from .interfaces import (
    bundle as i_bundle,
    exchange as i_exch,
    repository as i_repo,
)

if typing.TYPE_CHECKING:
    from typing import (
        Iterator,
    )


shape_mod = policy.importrust("shape")

urlerr = util.urlerr
urlreq = util.urlreq

_pack = struct.pack
_unpack = struct.unpack

_fstreamparamsize = b'>i'
_fpartheadersize = b'>i'
_fparttypesize = b'>B'
_fpartid = b'>I'
_fpayloadsize = b'>i'
_fpartparamcount = b'>BB'

preferedchunksize = 32768

_parttypeforbidden = re.compile(b'[^a-zA-Z0-9_:-]')


rbcstruct = struct.Struct(b'>III')


def outdebug(ui, message):
    """debug regarding output stream (bundling)"""
    if ui.configbool(b'devel', b'bundle2.debug'):
        ui.debug(b'bundle2-output: %s\n' % message)


def indebug(ui, message):
    """debug on input stream (unbundling)"""
    if ui.configbool(b'devel', b'bundle2.debug'):
        ui.debug(b'bundle2-input: %s\n' % message)


def validateparttype(parttype):
    """raise ValueError if a parttype contains invalid character"""
    if _parttypeforbidden.search(parttype):
        raise ValueError(parttype)


def _makefpartparamsizes(nbparams):
    """return a struct format to read part parameter sizes

    The number parameters is variable so we need to build that format
    dynamically.
    """
    return b'>' + (b'BB' * nbparams)


parthandlermapping = tables.bundle2_part_handler_mapping


class unbundlerecords(i_exch.IUnbundleRecords):
    """keep record of what happens during and unbundle

    New records are added using `records.add('cat', obj)`. Where 'cat' is a
    category of record and obj is an arbitrary object.

    `records['cat']` will return all entries of this category 'cat'.

    Iterating on the object itself will yield `('category', obj)` tuples
    for all entries.

    All iterations happens in chronological order.
    """

    def __init__(self):
        self._categories = {}
        self._sequences = []
        self._replies = {}

    def add(self, category, entry, inreplyto=None):
        """add a new record of a given category.

        The entry can then be retrieved in the list returned by
        self['category']."""
        self._categories.setdefault(category, []).append(entry)
        self._sequences.append((category, entry))
        if inreplyto is not None:
            self.getreplies(inreplyto).add(category, entry)

    def getreplies(self, partid):
        """get the records that are replies to a specific part"""
        return self._replies.setdefault(partid, unbundlerecords())

    def __getitem__(self, cat):
        return tuple(self._categories.get(cat, ()))

    def __iter__(self):
        return iter(self._sequences)

    def __len__(self):
        return len(self._sequences)

    def __nonzero__(self):
        return bool(self._sequences)

    __bool__ = __nonzero__


class bundleoperation(i_exch.IUnbundleOperation):
    """an object that represents a single bundling process

    Its purpose is to carry unbundle-related objects and states.

    A new object should be created at the beginning of each bundle processing.
    The object is to be returned by the processing function.

    The object has very little content now it will ultimately contain:
    * an access to the repo the bundle is applied to,
    * a ui object,
    * a way to retrieve a transaction to add changes to the repo,
    * a way to record the result of processing each part,
    * a way to construct a bundle response when applicable.
    """

    def __init__(
        self,
        repo,
        transactiongetter,
        captureoutput=True,
        source=b'',
        remote=None,
    ):
        self.repo = repo
        # the peer object who produced this bundle if available
        self.remote = remote
        self.ui = repo.ui
        self.records = unbundlerecords()
        self.reply = None
        self.captureoutput = captureoutput
        self.hookargs = {}
        self._gettransaction = transactiongetter
        # carries value that can modify part behavior
        self.modes = {}
        self.source = source

    def gettransaction(self):
        transaction = self._gettransaction()

        if self.hookargs:
            # the ones added to the transaction supercede those added
            # to the operation.
            self.hookargs.update(transaction.hookargs)
            transaction.hookargs = self.hookargs

        # mark the hookargs as flushed.  further attempts to add to
        # hookargs will result in an abort.
        self.hookargs = None

        return transaction

    def addhookargs(self, hookargs):
        if self.hookargs is None:
            raise error.ProgrammingError(
                b'attempted to add hookargs to '
                b'operation after transaction started'
            )
        self.hookargs.update(hookargs)


class TransactionUnavailable(RuntimeError):
    pass


def _notransaction():
    """default method to get a transaction while processing a bundle

    Raise an exception to highlight the fact that no transaction was expected
    to be created"""
    raise TransactionUnavailable()


def applybundle(repo, unbundler, tr, source, url=None, remote=None, **kwargs):
    # transform me into unbundler.apply() as soon as the freeze is lifted
    if isinstance(unbundler, unbundle20):
        tr.hookargs[b'bundle2'] = b'1'
        if source is not None and b'source' not in tr.hookargs:
            tr.hookargs[b'source'] = source
        if url is not None and b'url' not in tr.hookargs:
            tr.hookargs[b'url'] = url
        return processbundle(
            repo, unbundler, lambda: tr, source=source, remote=remote
        )
    else:
        # the transactiongetter won't be used, but we might as well set it
        op = bundleoperation(repo, lambda: tr, source=source, remote=remote)
        process_changegroup(op, unbundler, tr, source, url, **kwargs)
        return op


class partiterator:
    def __init__(self, repo, op, unbundler):
        self.repo = repo
        self.op = op
        self.unbundler = unbundler
        self.iterator = None
        self.count = 0
        self.current = None

    def __enter__(self):
        def func():
            itr = enumerate(self.unbundler.iterparts(), 1)
            for count, p in itr:
                self.count = count
                self.current = p
                yield p
                p.consume()
                self.current = None

        self.iterator = func()
        return self.iterator

    def __exit__(self, type, exc, tb):
        if not self.iterator:
            return

        # Only gracefully abort in a normal exception situation. User aborts
        # like Ctrl+C throw a KeyboardInterrupt which is not a base Exception,
        # and should not gracefully cleanup.
        if isinstance(exc, Exception):
            # Any exceptions seeking to the end of the bundle at this point are
            # almost certainly related to the underlying stream being bad.
            # And, chances are that the exception we're handling is related to
            # getting in that bad state. So, we swallow the seeking error and
            # re-raise the original error.
            seekerror = False
            try:
                if self.current:
                    # consume the part content to not corrupt the stream.
                    self.current.consume()

                for part in self.iterator:
                    # consume the bundle content
                    part.consume()
            except Exception:
                seekerror = True

            # Small hack to let caller code distinguish exceptions from bundle2
            # processing from processing the old format. This is mostly needed
            # to handle different return codes to unbundle according to the type
            # of bundle. We should probably clean up or drop this return code
            # craziness in a future version.
            exc.duringunbundle2 = True
            salvaged = []
            replycaps = None
            if self.op.reply is not None:
                salvaged = self.op.reply.salvageoutput()
                replycaps = self.op.reply.capabilities
            exc._replycaps = replycaps
            exc._bundle2salvagedoutput = salvaged

            # Re-raising from a variable loses the original stack. So only use
            # that form if we need to.
            if seekerror:
                raise exc

        self.repo.ui.debug(
            b'bundle2-input-bundle: %i parts total\n' % self.count
        )


def processbundle(
    repo,
    unbundler,
    transactiongetter=None,
    op=None,
    source=b'',
    remote=None,
):
    """This function process a bundle, apply effect to/from a repo

    It iterates over each part then searches for and uses the proper handling
    code to process the part. Parts are processed in order.

    Unknown Mandatory part will abort the process.

    It is temporarily possible to provide a prebuilt bundleoperation to the
    function. This is used to ensure output is properly propagated in case of
    an error during the unbundling. This output capturing part will likely be
    reworked and this ability will probably go away in the process.
    """
    if op is None:
        if transactiongetter is None:
            transactiongetter = _notransaction
        op = bundleoperation(
            repo,
            transactiongetter,
            source=source,
            remote=remote,
        )
    # todo:
    # - replace this is a init function soon.
    # - exception catching
    unbundler.params
    if repo.ui.debugflag:
        msg = [b'bundle2-input-bundle:']
        if unbundler.params:
            msg.append(b' %i params' % len(unbundler.params))
        if op._gettransaction is None or op._gettransaction is _notransaction:
            msg.append(b' no-transaction')
        else:
            msg.append(b' with-transaction')
        msg.append(b'\n')
        repo.ui.debug(b''.join(msg))

    processparts(repo, op, unbundler)

    return op


def processparts(repo: RepoT, op: UnbundleOpT, unbundler):
    with partiterator(repo, op, unbundler) as parts:
        for part in parts:
            _processpart(op, part)


def process_changegroup(op: UnbundleOpT, cg, tr, source, url, **kwargs):
    if op.remote is not None and op.remote.path is not None:
        remote_path = op.remote.path
        kwargs = kwargs.copy()
        kwargs['delta_base_reuse_policy'] = remote_path.delta_reuse_policy
    ret = cg.apply(op.repo, tr, source, url, **kwargs)
    op.records.add(
        b'changegroup',
        {
            b'return': ret,
        },
    )
    return ret


def _gethandler(op: UnbundleOpT | interruptoperation, part):
    status = b'unknown'  # used by debug output
    try:
        handler = parthandlermapping.get(part.type)
        if handler is None:
            status = b'unsupported-type'
            raise error.BundleUnknownFeatureError(parttype=part.type)
        indebug(op.ui, b'found a handler for part %s' % part.type)
        unknownparams = part.mandatorykeys - handler.params
        if unknownparams:
            unknownparams = list(unknownparams)
            unknownparams.sort()
            status = b'unsupported-params (%s)' % b', '.join(unknownparams)
            raise error.BundleUnknownFeatureError(
                parttype=part.type, params=unknownparams
            )
        status = b'supported'
    except error.BundleUnknownFeatureError as exc:
        if part.mandatory:  # mandatory parts
            raise
        indebug(op.ui, b'ignoring unsupported advisory part %s' % exc)
        return  # skip to part processing
    finally:
        if op.ui.debugflag:
            msg = [b'bundle2-input-part: "%s"' % part.type]
            if not part.mandatory:
                msg.append(b' (advisory)')
            nbmp = len(part.mandatorykeys)
            nbap = len(part.params) - nbmp
            if nbmp or nbap:
                msg.append(b' (params:')
                if nbmp:
                    msg.append(b' %i mandatory' % nbmp)
                if nbap:
                    msg.append(b' %i advisory' % nbmp)
                msg.append(b')')
            msg.append(b' %s\n' % status)
            op.ui.debug(b''.join(msg))

    return handler


def _processpart(op: UnbundleOpT | interruptoperation, part):
    """process a single part from a bundle

    The part is guaranteed to have been fully consumed when the function exits
    (even if an exception is raised)."""
    handler = _gethandler(op, part)
    if handler is None:
        return

    # handler is called outside the above try block so that we don't
    # risk catching KeyErrors from anything other than the
    # parthandlermapping lookup (any KeyError raised by handler()
    # itself represents a defect of a different variety).
    output = None
    if op.captureoutput and op.reply is not None:
        op.ui.pushbuffer(error=True, subproc=True)
        output = b''
    try:
        handler(op, part)
    finally:
        if output is not None:
            output = op.ui.popbuffer()
        if output:
            # The `is None` case is tested early and won't walk in this branch.
            # the assert helps pytype
            reply = op.reply
            assert reply is not None
            outpart = reply.newpart(b'output', data=output, mandatory=False)
            outpart.addparam(
                b'in-reply-to', pycompat.bytestr(part.id), mandatory=False
            )


decodecaps = urlutil.decode_b2_caps


def encodecaps(caps):
    """encode a bundle2 caps dictionary into a bytes blob"""
    chunks = []
    for ca in sorted(caps):
        vals = caps[ca]
        ca = urlreq.quote(ca)
        vals = [urlreq.quote(v) for v in vals]
        if vals:
            ca = b"%s=%s" % (ca, b','.join(vals))
        chunks.append(ca)
    return b'\n'.join(chunks)


bundletypes = {
    b"": (b"", b'UN'),  # only when using unbundle on ssh and old http servers
    # since the unification ssh accepts a header but there
    # is no capability signaling it.
    b"HG20": (),  # special-cased below
    b"HG10UN": (b"HG10UN", b'UN'),
    b"HG10BZ": (b"HG10", b'BZ'),
    b"HG10GZ": (b"HG10GZ", b'GZ'),
}

# hgweb uses this list to communicate its preferred type
bundlepriority = [b'HG10GZ', b'HG10BZ', b'HG10UN']


class bundle20(i_exch.IBundle20):
    """represent an outgoing bundle2 container

    Use the `addparam` method to add stream level parameter. and `newpart` to
    populate it. Then call `getchunks` to retrieve all the binary chunks of
    data that compose the bundle2 container."""

    _magicstring = b'HG20'

    def __init__(self, ui, capabilities: Capabilities | None = None):
        if capabilities is None:
            capabilities = {}

        self.ui = ui
        self._params = []
        self._parts = []
        self.capabilities: Capabilities = dict(capabilities)
        self._compengine = util.compengines.forbundletype(b'UN')
        self._compopts = None
        # If compression is being handled by a consumer of the raw
        # data (e.g. the wire protocol), unsetting this flag tells
        # consumers that the bundle is best left uncompressed.
        self.prefercompressed = True

    def setcompression(self, alg, compopts=None):
        """setup core part compression to <alg>"""
        if alg in (None, b'UN'):
            return
        assert not any(n.lower() == b'compression' for n, v in self._params)
        self.addparam(b'Compression', alg)
        self._compengine = util.compengines.forbundletype(alg)
        self._compopts = compopts

    @property
    def nbparts(self):
        """total number of parts added to the bundler"""
        return len(self._parts)

    # methods used to defines the bundle2 content
    def addparam(self, name, value=None):
        """add a stream level parameter"""
        if not name:
            raise error.ProgrammingError(b'empty parameter name')
        if name[0:1] not in pycompat.bytestr(
            string.ascii_letters  # pytype: disable=wrong-arg-types
        ):
            raise error.ProgrammingError(
                b'non letter first character: %s' % name
            )
        self._params.append((name, value))

    def addpart(self, part):
        """add a new part to the bundle2 container

        Parts contains the actual applicative payload."""
        assert part.id is None
        part.id = len(self._parts)  # very cheap counter
        self._parts.append(part)

    def newpart(self, typeid, *args, **kwargs):
        """create a new part and add it to the containers

        As the part is directly added to the containers. For now, this means
        that any failure to properly initialize the part after calling
        ``newpart`` should result in a failure of the whole bundling process.

        You can still fall back to manually create and add if you need better
        control."""
        part = bundlepart(typeid, *args, **kwargs)
        self.addpart(part)
        return part

    # methods used to generate the bundle2 stream
    def getchunks(self) -> Iterator[bytes]:
        if self.ui.debugflag:
            msg = [b'bundle2-output-bundle: "%s",' % self._magicstring]
            if self._params:
                msg.append(b' (%i params)' % len(self._params))
            msg.append(b' %i parts total\n' % len(self._parts))
            self.ui.debug(b''.join(msg))
        outdebug(self.ui, b'start emission of %s stream' % self._magicstring)
        yield self._magicstring
        param = self._paramchunk()
        outdebug(self.ui, b'bundle parameter: %s' % param)
        yield _pack(_fstreamparamsize, len(param))
        if param:
            yield param
        yield from self._compengine.compressstream(
            self._getcorechunk(), self._compopts
        )

    def _paramchunk(self):
        """return a encoded version of all stream parameters"""
        blocks = []
        for par, value in self._params:
            par = urlreq.quote(par)
            if value is not None:
                value = urlreq.quote(value)
                par = b'%s=%s' % (par, value)
            blocks.append(par)
        return b' '.join(blocks)

    def _getcorechunk(self):
        """yield chunk for the core part of the bundle

        (all but headers and parameters)"""
        outdebug(self.ui, b'start of parts')
        for part in self._parts:
            outdebug(self.ui, b'bundle part: "%s"' % part.type)
            yield from part.getchunks(ui=self.ui)
        outdebug(self.ui, b'end of bundle')
        yield _pack(_fpartheadersize, 0)

    def salvageoutput(self):
        """return a list with a copy of all output parts in the bundle

        This is meant to be used during error handling to make sure we preserve
        server output"""
        salvaged = []
        for part in self._parts:
            if part.type.startswith(b'output'):
                salvaged.append(part.copy())
        return salvaged


class unpackermixin:
    """A mixin to extract bytes and struct data from a stream"""

    def __init__(self, fp):
        self._fp = fp

    def _unpack(self, format):
        """unpack this struct format from the stream

        This method is meant for internal usage by the bundle2 protocol only.
        They directly manipulate the low level stream including bundle2 level
        instruction.

        Do not use it to implement higher-level logic or methods."""
        data = self._readexact(struct.calcsize(format))
        return _unpack(format, data)

    def _readexact(self, size):
        """read exactly <size> bytes from the stream

        This method is meant for internal usage by the bundle2 protocol only.
        They directly manipulate the low level stream including bundle2 level
        instruction.

        Do not use it to implement higher-level logic or methods."""
        return changegroup.readexactly(self._fp, size)


def getunbundler(ui, fp, magicstring=None):
    """return a valid unbundler object for a given magicstring"""
    if magicstring is None:
        magicstring = changegroup.readexactly(fp, 4)
    magic, version = magicstring[0:2], magicstring[2:4]
    if magic != b'HG':
        ui.debug(
            b"error: invalid magic: %r (version %r), should be 'HG'\n"
            % (magic, version)
        )
        raise error.Abort(_(b'not a Mercurial bundle'))
    unbundlerclass = formatmap.get(version)
    if unbundlerclass is None:
        raise error.Abort(_(b'unknown bundle version %s') % version)
    unbundler = unbundlerclass(ui, fp)
    indebug(ui, b'start processing of %s stream' % magicstring)
    return unbundler


class unbundle20(unpackermixin):
    """interpret a bundle2 stream

    This class is fed with a binary stream and yields parts through its
    `iterparts` methods."""

    _magicstring = b'HG20'

    def __init__(self, ui, fp):
        """If header is specified, we do not read it out of the stream."""
        self.ui = ui
        self._compengine = util.compengines.forbundletype(b'UN')
        self._compressed = None
        super().__init__(fp)

    @util.propertycache
    def params(self):
        """dictionary of stream level parameters"""
        indebug(self.ui, b'reading bundle2 stream parameters')
        params = {}
        paramssize = self._unpack(_fstreamparamsize)[0]
        if paramssize < 0:
            raise error.BundleValueError(
                b'negative bundle param size: %i' % paramssize
            )
        if paramssize:
            params = self._readexact(paramssize)
            params = self._processallparams(params)
        return params

    def _processallparams(
        self, paramsblock: bytes
    ) -> util.sortdict[bytes, bytes]:
        """ """
        params = util.sortdict()
        for p in paramsblock.split(b' '):
            p = p.split(b'=', 1)
            p = [urlreq.unquote(i) for i in p]
            if len(p) < 2:
                p.append(None)
            self._processparam(*p)
            params[p[0]] = p[1]
        return params

    def _processparam(self, name, value):
        """process a parameter, applying its effect if needed

        Parameter starting with a lower case letter are advisory and will be
        ignored when unknown.  Those starting with an upper case letter are
        mandatory and will this function will raise a KeyError when unknown.

        Note: no option are currently supported. Any input will be either
              ignored or failing.
        """
        if not name:
            raise ValueError('empty parameter name')
        if name[0:1] not in pycompat.bytestr(
            string.ascii_letters  # pytype: disable=wrong-arg-types
        ):
            raise ValueError('non letter first character: %s' % name)
        try:
            handler = b2streamparamsmap[name.lower()]
        except KeyError:
            if name[0:1].islower():
                indebug(self.ui, b"ignoring unknown parameter %s" % name)
            else:
                raise error.BundleUnknownFeatureError(params=(name,))
        else:
            handler(self, name, value)

    def _forwardchunks(self):
        """utility to transfer a bundle2 as binary

        This is made necessary by the fact the 'getbundle' command over 'ssh'
        have no way to know when the reply ends, relying on the bundle to be
        interpreted to know its end. This is terrible and we are sorry, but we
        needed to move forward to get general delta enabled.
        """
        yield self._magicstring
        assert 'params' not in vars(self)
        paramssize = self._unpack(_fstreamparamsize)[0]
        if paramssize < 0:
            raise error.BundleValueError(
                b'negative bundle param size: %i' % paramssize
            )
        if paramssize:
            params = self._readexact(paramssize)
            self._processallparams(params)
            # The payload itself is decompressed below, so drop
            # the compression parameter passed down to compensate.
            outparams = []
            for p in params.split(b' '):
                k, v = p.split(b'=', 1)
                if k.lower() != b'compression':
                    outparams.append(p)
            outparams = b' '.join(outparams)
            yield _pack(_fstreamparamsize, len(outparams))
            yield outparams
        else:
            yield _pack(_fstreamparamsize, paramssize)
        # From there, payload might need to be decompressed
        self._fp = self._compengine.decompressorreader(self._fp)

        # We usually wait for empty terminators: a bundle part terminator
        # followed by a bundle terminator.
        #
        # Since the empty bundle has no parts,
        # bundle part terminator never comes.
        #
        # So we boostrap this detection with am empty part, which the first
        # piece of part will reset.
        emptycount = 1
        while emptycount < 2:
            # so we can brainlessly loop
            assert _fpartheadersize == _fpayloadsize
            size = self._unpack(_fpartheadersize)[0]
            yield _pack(_fpartheadersize, size)
            if size:
                emptycount = 0
            else:
                emptycount += 1
                continue
            if size == flaginterrupt:
                continue
            elif size < 0:
                raise error.BundleValueError(b'negative chunk size: %i')
            yield self._readexact(size)

    def iterparts(self, seekable=False):
        """yield all parts contained in the stream"""
        cls = seekableunbundlepart if seekable else unbundlepart
        # make sure param have been loaded
        self.params
        # From there, payload need to be decompressed
        self._fp = self._compengine.decompressorreader(self._fp)
        indebug(self.ui, b'start extraction of bundle2 parts')
        headerblock = self._readpartheader()
        while headerblock is not None:
            part = cls(self.ui, headerblock, self._fp)
            yield part
            # Ensure part is fully consumed so we can start reading the next
            # part.
            part.consume()

            headerblock = self._readpartheader()
        indebug(self.ui, b'end of bundle2 stream')

    def _readpartheader(self):
        """reads a part header size and return the bytes blob

        returns None if empty"""
        headersize = self._unpack(_fpartheadersize)[0]
        if headersize < 0:
            raise error.BundleValueError(
                b'negative part header size: %i' % headersize
            )
        indebug(self.ui, b'part header size: %i' % headersize)
        if headersize:
            return self._readexact(headersize)
        return None

    def compressed(self):
        self.params  # load params
        return self._compressed

    def close(self):
        """close underlying file"""
        if hasattr(self._fp, 'close'):
            return self._fp.close()


formatmap = {b'20': unbundle20}

b2streamparamsmap = {}


def b2streamparamhandler(name):
    """register a handler for a stream level parameter"""

    def decorator(func):
        assert name not in formatmap
        b2streamparamsmap[name] = func
        return func

    return decorator


@b2streamparamhandler(b'compression')
def processcompression(unbundler, param, value):
    """read compression parameter and install payload decompression"""
    if value not in util.compengines.supportedbundletypes:
        raise error.BundleUnknownFeatureError(params=(param,), values=(value,))
    unbundler._compengine = util.compengines.forbundletype(value)
    if value is not None:
        unbundler._compressed = True


class bundlepart(i_exch.IBundlePart):
    """A bundle2 part contains application level payload

    The part `type` is used to route the part to the application level
    handler.

    The part payload is contained in ``part.data``. It could be raw bytes or a
    generator of byte chunks.

    You can add parameters to the part using the ``addparam`` method.
    Parameters can be either mandatory (default) or advisory. Remote side
    should be able to safely ignore the advisory ones.

    Both data and parameters cannot be modified after the generation has begun.
    """

    def __init__(
        self,
        parttype,
        mandatoryparams=(),
        advisoryparams=(),
        data=b'',
        mandatory=True,
    ):
        validateparttype(parttype)
        self.id = None
        self.type = parttype
        self._data = data
        self._mandatoryparams = list(mandatoryparams)
        self._advisoryparams = list(advisoryparams)
        # checking for duplicated entries
        self._seenparams = set()
        for pname, __ in self._mandatoryparams + self._advisoryparams:
            if pname in self._seenparams:
                raise error.ProgrammingError(b'duplicated params: %s' % pname)
            self._seenparams.add(pname)
        # status of the part's generation:
        # - None: not started,
        # - False: currently generated,
        # - True: generation done.
        self._generated = None
        self.mandatory = mandatory

    def __repr__(self):
        cls = "%s.%s" % (self.__class__.__module__, self.__class__.__name__)
        return '<%s object at %x; id: %s; type: %s; mandatory: %s>' % (
            cls,
            id(self),
            self.id,
            self.type,
            self.mandatory,
        )

    def copy(self):
        """return a copy of the part

        The new part have the very same content but no partid assigned yet.
        Parts with generated data cannot be copied."""
        assert not hasattr(self.data, 'next')
        return self.__class__(
            self.type,
            self._mandatoryparams,
            self._advisoryparams,
            self._data,
            self.mandatory,
        )

    # methods used to defines the part content
    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, data):
        if self._generated is not None:
            raise error.ReadOnlyPartError(b'part is being generated')
        self._data = data

    @property
    def mandatoryparams(self):
        # make it an immutable tuple to force people through ``addparam``
        return tuple(self._mandatoryparams)

    @property
    def advisoryparams(self):
        # make it an immutable tuple to force people through ``addparam``
        return tuple(self._advisoryparams)

    def addparam(self, name, value=b'', mandatory=True):
        """add a parameter to the part

        If 'mandatory' is set to True, the remote handler must claim support
        for this parameter or the unbundling will be aborted.

        The 'name' and 'value' cannot exceed 255 bytes each.
        """
        if self._generated is not None:
            raise error.ReadOnlyPartError(b'part is being generated')
        if name in self._seenparams:
            raise ValueError(b'duplicated params: %s' % name)
        self._seenparams.add(name)
        params = self._advisoryparams
        if mandatory:
            params = self._mandatoryparams
        params.append((name, value))

    # methods used to generates the bundle2 stream
    def getchunks(self, ui):
        if self._generated is not None:
            raise error.ProgrammingError(b'part can only be consumed once')
        self._generated = False

        if ui.debugflag:
            msg = [b'bundle2-output-part: "%s"' % self.type]
            if not self.mandatory:
                msg.append(b' (advisory)')
            nbmp = len(self.mandatoryparams)
            nbap = len(self.advisoryparams)
            if nbmp or nbap:
                msg.append(b' (params:')
                if nbmp:
                    msg.append(b' %i mandatory' % nbmp)
                if nbap:
                    msg.append(b' %i advisory' % nbmp)
                msg.append(b')')
            if not self.data:
                msg.append(b' empty payload')
            elif hasattr(self.data, 'next') or hasattr(self.data, '__next__'):
                msg.append(b' streamed payload')
            else:
                msg.append(b' %i bytes payload' % len(self.data))
            msg.append(b'\n')
            ui.debug(b''.join(msg))

        #### header
        if self.mandatory:
            parttype = self.type.upper()
        else:
            parttype = self.type.lower()
        outdebug(ui, b'part %s: "%s"' % (pycompat.bytestr(self.id), parttype))
        ## parttype
        header = [
            _pack(_fparttypesize, len(parttype)),
            parttype,
            _pack(_fpartid, self.id),
        ]
        ## parameters
        # count
        manpar = self.mandatoryparams
        advpar = self.advisoryparams
        header.append(_pack(_fpartparamcount, len(manpar), len(advpar)))
        # size
        parsizes = []
        for key, value in manpar:
            parsizes.append(len(key))
            parsizes.append(len(value))
        for key, value in advpar:
            parsizes.append(len(key))
            parsizes.append(len(value))
        paramsizes = _pack(_makefpartparamsizes(len(parsizes) // 2), *parsizes)
        header.append(paramsizes)
        # key, value
        for key, value in manpar:
            header.append(key)
            header.append(value)
        for key, value in advpar:
            header.append(key)
            header.append(value)
        ## finalize header
        try:
            headerchunk = b''.join(header)
        except TypeError:
            raise TypeError(
                'Found a non-bytes trying to '
                'build bundle part header: %r' % header
            )
        outdebug(ui, b'header chunk size: %i' % len(headerchunk))
        yield _pack(_fpartheadersize, len(headerchunk))
        yield headerchunk
        ## payload
        try:
            for chunk in self._payloadchunks():
                outdebug(ui, b'payload chunk size: %i' % len(chunk))
                yield _pack(_fpayloadsize, len(chunk))
                yield chunk
        except GeneratorExit:
            # GeneratorExit means that nobody is listening for our
            # results anyway, so just bail quickly rather than trying
            # to produce an error part.
            ui.debug(b'bundle2-generatorexit\n')
            raise
        except BaseException as exc:
            bexc = stringutil.forcebytestr(exc)
            # backup exception data for later
            ui.debug(
                b'bundle2-input-stream-interrupt: encoding exception %s' % bexc
            )
            tb = sys.exc_info()[2]
            msg = b'unexpected error: %s' % bexc
            interpart = bundlepart(
                b'error:abort', [(b'message', msg)], mandatory=False
            )
            interpart.id = 0
            yield _pack(_fpayloadsize, -1)
            for chunk in interpart.getchunks(ui=ui):
                yield chunk
            outdebug(ui, b'closing payload chunk')
            # abort current part payload
            yield _pack(_fpayloadsize, 0)
            pycompat.raisewithtb(exc, tb)
        # end of payload
        outdebug(ui, b'closing payload chunk')
        yield _pack(_fpayloadsize, 0)
        self._generated = True

    def _payloadchunks(self):
        """yield chunks of a the part payload

        Exists to handle the different methods to provide data to a part."""
        # we only support fixed size data now.
        # This will be improved in the future.
        if hasattr(self.data, 'next') or hasattr(self.data, '__next__'):
            buff = util.chunkbuffer(self.data)
            chunk = buff.read(preferedchunksize)
            while chunk:
                yield chunk
                chunk = buff.read(preferedchunksize)
        elif len(self.data):
            yield self.data


flaginterrupt = -1


class interrupthandler(unpackermixin):
    """read one part and process it with restricted capability

    This allows to transmit exception raised on the producer size during part
    iteration while the consumer is reading a part.

    Part processed in this manner only have access to a ui object,"""

    def __init__(self, ui, fp):
        super().__init__(fp)
        self.ui = ui

    def _readpartheader(self):
        """reads a part header size and return the bytes blob

        returns None if empty"""
        headersize = self._unpack(_fpartheadersize)[0]
        if headersize < 0:
            raise error.BundleValueError(
                b'negative part header size: %i' % headersize
            )
        indebug(self.ui, b'part header size: %i\n' % headersize)
        if headersize:
            return self._readexact(headersize)
        return None

    def __call__(self):
        self.ui.debug(
            b'bundle2-input-stream-interrupt: opening out of band context\n'
        )
        indebug(self.ui, b'bundle2 stream interruption, looking for a part.')
        headerblock = self._readpartheader()
        if headerblock is None:
            indebug(self.ui, b'no part found during interruption.')
            return
        part = unbundlepart(self.ui, headerblock, self._fp)
        op = interruptoperation(self.ui)
        hardabort = False
        try:
            _processpart(op, part)
        except (SystemExit, KeyboardInterrupt):
            hardabort = True
            raise
        finally:
            if not hardabort:
                part.consume()
        self.ui.debug(
            b'bundle2-input-stream-interrupt: closing out of band context\n'
        )


class interruptoperation:
    """A limited operation to be use by part handler during interruption

    It only have access to an ui object.
    """

    def __init__(self, ui):
        self.ui = ui
        self.reply: None = None
        self.captureoutput = False

    @property
    def repo(self):
        raise error.ProgrammingError(b'no repo access from stream interruption')

    def gettransaction(self):
        raise TransactionUnavailable(b'no repo access from stream interruption')


def decodepayloadchunks(ui, fh):
    """Reads bundle2 part payload data into chunks.

    Part payload data consists of framed chunks. This function takes
    a file handle and emits those chunks.
    """
    dolog = ui.configbool(b'devel', b'bundle2.debug')
    debug = ui.debug

    headerstruct = struct.Struct(_fpayloadsize)
    headersize = headerstruct.size
    unpack = headerstruct.unpack

    readexactly = changegroup.readexactly
    read = fh.read

    chunksize = unpack(readexactly(fh, headersize))[0]
    indebug(ui, b'payload chunk size: %i' % chunksize)

    # changegroup.readexactly() is inlined below for performance.
    while chunksize:
        if chunksize >= 0:
            s = read(chunksize)
            if len(s) < chunksize:
                raise error.Abort(
                    _(
                        b'stream ended unexpectedly '
                        b' (got %d bytes, expected %d)'
                    )
                    % (len(s), chunksize)
                )

            yield s
        elif chunksize == flaginterrupt:
            # Interrupt "signal" detected. The regular stream is interrupted
            # and a bundle2 part follows. Consume it.
            interrupthandler(ui, fh)()
        else:
            raise error.BundleValueError(
                b'negative payload chunk size: %s' % chunksize
            )

        s = read(headersize)
        if len(s) < headersize:
            raise error.Abort(
                _(b'stream ended unexpectedly  (got %d bytes, expected %d)')
                % (len(s), chunksize)
            )

        chunksize = unpack(s)[0]

        # indebug() inlined for performance.
        if dolog:
            debug(b'bundle2-input: payload chunk size: %i\n' % chunksize)


class unbundlepart(unpackermixin, i_bundle.IUnbundlePart):
    """a bundle part read from a bundle"""

    def __init__(self, ui, header, fp):
        super().__init__(fp)
        self._seekable = hasattr(fp, 'seek') and hasattr(fp, 'tell')
        self.ui = ui
        # unbundle state attr
        self._headerdata = header
        self._headeroffset = 0
        self._initialized = False
        self.consumed = False
        # part data
        self.id = None
        self.type = None
        self.mandatoryparams = None
        self.advisoryparams = None
        self.params = None
        self.mandatorykeys = ()
        self._readheader()
        self._mandatory = None
        self._pos = 0

    def _fromheader(self, size):
        """return the next <size> byte from the header"""
        offset = self._headeroffset
        data = self._headerdata[offset : (offset + size)]
        self._headeroffset = offset + size
        return data

    def _unpackheader(self, format):
        """read given format from header

        This automatically compute the size of the format to read."""
        data = self._fromheader(struct.calcsize(format))
        return _unpack(format, data)

    def _initparams(self, mandatoryparams, advisoryparams):
        """internal function to setup all logic related parameters"""
        # make it read only to prevent people touching it by mistake.
        self.mandatoryparams = tuple(mandatoryparams)
        self.advisoryparams = tuple(advisoryparams)
        # user friendly UI
        self.params = util.sortdict(self.mandatoryparams)
        self.params.update(self.advisoryparams)
        self.mandatorykeys = frozenset(p[0] for p in mandatoryparams)

    def _readheader(self):
        """read the header and setup the object"""
        typesize = self._unpackheader(_fparttypesize)[0]
        self.type = self._fromheader(typesize)
        indebug(self.ui, b'part type: "%s"' % self.type)
        self.id = self._unpackheader(_fpartid)[0]
        indebug(self.ui, b'part id: "%s"' % pycompat.bytestr(self.id))
        # extract mandatory bit from type
        self.mandatory = self.type != self.type.lower()
        self.type = self.type.lower()
        ## reading parameters
        # param count
        mancount, advcount = self._unpackheader(_fpartparamcount)
        indebug(self.ui, b'part parameters: %i' % (mancount + advcount))
        # param size
        fparamsizes = _makefpartparamsizes(mancount + advcount)
        paramsizes = self._unpackheader(fparamsizes)
        # make it a list of couple again
        paramsizes = list(zip(paramsizes[::2], paramsizes[1::2]))
        # split mandatory from advisory
        mansizes = paramsizes[:mancount]
        advsizes = paramsizes[mancount:]
        # retrieve param value
        manparams = []
        for key, value in mansizes:
            manparams.append((self._fromheader(key), self._fromheader(value)))
        advparams = []
        for key, value in advsizes:
            advparams.append((self._fromheader(key), self._fromheader(value)))
        self._initparams(manparams, advparams)
        ## part payload
        self._payloadstream = util.chunkbuffer(self._payloadchunks())
        # we read the data, tell it
        self._initialized = True

    def __iter__(self):
        for chunk in self._payloadstream:
            self._pos += len(chunk)
            yield chunk

    def _payloadchunks(self):
        """Generator of decoded chunks in the payload."""
        return decodepayloadchunks(self.ui, self._fp)

    def consume(self):
        """Read the part payload until completion.

        By consuming the part data, the underlying stream read offset will
        be advanced to the next part (or end of stream).
        """
        if self.consumed:
            return

        chunk = self.read(32768)
        while chunk:
            self._pos += len(chunk)
            chunk = self.read(32768)

    def read(self, size=None):
        """read payload data"""
        if not self._initialized:
            self._readheader()
        if size is None:
            data = self._payloadstream.read()
        else:
            data = self._payloadstream.read(size)
        self._pos += len(data)
        if size is None or len(data) < size:
            if not self.consumed and self._pos:
                self.ui.debug(
                    b'bundle2-input-part: total payload size %i\n' % self._pos
                )
            self.consumed = True
        return data

    def tell(self) -> int:
        """the amount of byte read so far in the part"""
        return self._payloadstream.tell()

    def as_seekable(self) -> seekableunbundlepart:
        raise error.ProgrammingError(b"part isn't from a seekable source")


class seekableunbundlepart(unbundlepart):
    """A bundle2 part in a bundle that is seekable.

    Regular ``unbundlepart`` instances can only be read once. This class
    extends ``unbundlepart`` to enable bi-directional seeking within the
    part.

    Bundle2 part data consists of framed chunks. Offsets when seeking
    refer to the decoded data, not the offsets in the underlying bundle2
    stream.

    To facilitate quickly seeking within the decoded data, instances of this
    class maintain a mapping between offsets in the underlying stream and
    the decoded payload. This mapping will consume memory in proportion
    to the number of chunks within the payload (which almost certainly
    increases in proportion with the size of the part).
    """

    def __init__(self, ui, header, fp):
        # (payload, file) offsets for chunk starts.
        self._chunkindex = []

        super().__init__(ui, header, fp)

    def _payloadchunks(self, chunknum=0):
        '''seek to specified chunk and start yielding data'''
        if len(self._chunkindex) == 0:
            assert chunknum == 0, b'Must start with chunk 0'
            self._chunkindex.append((0, self._tellfp()))
        else:
            assert chunknum < len(self._chunkindex), (
                b'Unknown chunk %d' % chunknum
            )
            self._seekfp(self._chunkindex[chunknum][1])

        pos = self._chunkindex[chunknum][0]

        for chunk in decodepayloadchunks(self.ui, self._fp):
            chunknum += 1
            pos += len(chunk)
            if chunknum == len(self._chunkindex):
                self._chunkindex.append((pos, self._tellfp()))

            yield chunk

    def _findchunk(self, pos):
        '''for a given payload position, return a chunk number and offset'''
        for chunk, (ppos, fpos) in enumerate(self._chunkindex):
            if ppos == pos:
                return chunk, 0
            elif ppos > pos:
                return chunk - 1, pos - self._chunkindex[chunk - 1][0]
        raise ValueError(b'Unknown chunk')

    def tell(self):
        return self._pos

    def seek(self, offset, whence=os.SEEK_SET):
        if whence == os.SEEK_SET:
            newpos = offset
        elif whence == os.SEEK_CUR:
            newpos = self._pos + offset
        elif whence == os.SEEK_END:
            if not self.consumed:
                # Can't use self.consume() here because it advances self._pos.
                chunk = self.read(32768)
                while chunk:
                    chunk = self.read(32768)
            newpos = self._chunkindex[-1][0] - offset
        else:
            raise ValueError(b'Unknown whence value: %r' % (whence,))

        if newpos > self._chunkindex[-1][0] and not self.consumed:
            # Can't use self.consume() here because it advances self._pos.
            chunk = self.read(32768)
            while chunk:
                chunk = self.read(32668)

        if not 0 <= newpos <= self._chunkindex[-1][0]:
            raise ValueError(b'Offset out of range')

        if self._pos != newpos:
            chunk, internaloffset = self._findchunk(newpos)
            self._payloadstream = util.chunkbuffer(self._payloadchunks(chunk))
            adjust = self.read(internaloffset)
            if len(adjust) != internaloffset:
                raise error.Abort(_(b'Seek failed\n'))
            self._pos = newpos

    def _seekfp(self, offset, whence=0):
        """move the underlying file pointer

        This method is meant for internal usage by the bundle2 protocol only.
        They directly manipulate the low level stream including bundle2 level
        instruction.

        Do not use it to implement higher-level logic or methods."""
        if self._seekable:
            return self._fp.seek(offset, whence)
        else:
            raise NotImplementedError(_(b'File pointer is not seekable'))

    def _tellfp(self):
        """return the file offset, or None if file is not seekable

        This method is meant for internal usage by the bundle2 protocol only.
        They directly manipulate the low level stream including bundle2 level
        instruction.

        Do not use it to implement higher-level logic or methods."""
        if self._seekable:
            try:
                return self._fp.tell()
            except OSError as e:
                if e.errno == errno.ESPIPE:
                    self._seekable = False
                else:
                    raise
        return None

    def as_seekable(self) -> seekableunbundlepart:
        return self


# offered for compatibilities
capabilities: Capabilities = bundle_caps.capabilities

getrepocaps = util.deprecated(
    "Use `mercurial.exchanges.bundle_caps.get_repo_caps() instead`", "7.3"
)(bundle_caps.get_repo_caps)


def bundle2caps(remote) -> Capabilities:
    """return the bundle capabilities of a peer as dict"""
    raw = remote.capable(b'bundle2')
    if not raw and raw != b'':
        return {}
    capsblob = urlreq.unquote(remote.capable(b'bundle2'))
    return decodecaps(capsblob)


def obsmarkersversion(caps: Capabilities):
    """extract the list of supported obsmarkers versions from a bundle2caps dict"""
    obscaps = caps.get(b'obsmarkers', ())
    return [int(c[1:]) for c in obscaps if c.startswith(b'V')]


def writenewbundle(
    ui,
    repo,
    source,
    filename,
    bundletype,
    outgoing,
    opts,
    vfs=None,
    compression=None,
    compopts=None,
    allow_internal=False,
):
    if bundletype.startswith(b'HG10'):
        cg = changegroup.makechangegroup(repo, outgoing, b'01', source)
        return writebundle(
            ui,
            cg,
            filename,
            bundletype,
            vfs=vfs,
            compression=compression,
            compopts=compopts,
        )
    elif not bundletype.startswith(b'HG20'):
        raise error.ProgrammingError(b'unknown bundle type: %s' % bundletype)

    # enforce that no internal phase are to be bundled
    bundled_internal = repo.revs(b"%ln and _internal()", outgoing.ancestorsof)
    if bundled_internal and not allow_internal:
        count = len(repo.revs(b'%ln and _internal()', outgoing.missing))
        msg = "backup bundle would contains %d internal changesets"
        msg %= count
        raise error.ProgrammingError(msg)

    caps: Capabilities = {}
    if opts.get(b'obsolescence', False):
        caps[b'obsmarkers'] = (b'V1',)
    if stream_version := opts.get(b'stream', b""):
        return write_new_stream_bundle(
            repo=repo,
            version=stream_version,
            filename=filename,
            opts=opts,
            vfs=vfs,
            shape=opts.get(b"shape"),
        )
    else:
        bundle = bundle20(ui, caps)
        bundle.setcompression(compression, compopts)
        _addpartsfromopts(repo, bundle, source, outgoing, opts)
    chunkiter = bundle.getchunks()

    return changegroup.writechunks(ui, chunkiter, filename, vfs=vfs)


def write_new_stream_bundle(
    repo: RepoT,
    version: bytes,
    filename: bytes,
    opts,
    vfs: vfsmod.vfs,
    shape: bytes | None = None,
):
    ui = repo.ui

    matcher = None
    fingerprint = None

    if shape is not None:
        if shape_mod is None:
            raise error.ProgrammingError("shapes are a Rust-only feature")
        store_shards = shape_mod.get_store_shards(repo.root)
        shape_obj = store_shards.shape(shape.decode())
        if shape_obj is None:
            raise error.InputError(b"unknown shape: '%s'" % shape)
        matcher = shape_obj.matcher()
        fingerprint = shape_obj.fingerprint()

    caps: Capabilities = {b"stream": [version]}
    if opts.get(b'obsolescence', False):
        caps[b'obsmarkers'] = (b'V1',)

    bundle = bundle20(ui, caps)
    addpartbundlestream2(
        bundle,
        repo,
        narrow_matcher=matcher,
        stream=True,
        store_fingerprint=fingerprint,
    )

    return changegroup.writechunks(ui, bundle.getchunks(), filename, vfs=vfs)


def _addpartsfromopts(repo, bundler, source, outgoing, opts):
    # We should eventually reconcile this logic with the one behind
    # 'exchange.getbundle2partsgenerator'.
    #
    # The type of input from 'getbundle' and 'writenewbundle' are a bit
    # different right now. So we keep them separated for now for the sake of
    # simplicity.

    # we might not always want a changegroup in such bundle for legacy formats
    if opts.get(b'changegroup', True):
        cgversion = opts.get(b'cg.version')
        if cgversion is None:
            cgversion = changegroup.safeversion(repo)
        delta_comp = opts.get(b'cg.delta-compression')
        if delta_comp is not None:
            delta_comp = delta_comp.split(b',')

        bundlecaps = None
        if delta_comp is not None and cgversion in (b'04',):
            b2_caps = {b'delta-compression': delta_comp}
            bin_caps = encodecaps(b2_caps)
            bundlecaps = {b'bundle2=' + urlreq.quote(bin_caps)}
        cg = changegroup.makechangegroup(
            repo,
            outgoing,
            cgversion,
            source,
            bundlecaps=bundlecaps,
        )
        part = bundler.newpart(b'changegroup', data=cg.getchunks())
        part.addparam(b'version', cg.version)
        if b'clcount' in cg.extras:
            part.addparam(
                b'nbchanges', b'%d' % cg.extras[b'clcount'], mandatory=False
            )
        if b'delta-compression' in cg.extras:
            # Making the parameter advisory as bad compression will be caught
            # later down the road (I am not sure about it)
            part.addparam(
                b'delta-compression',
                b','.join(sorted(cg.extras[b'delta-compression'])),
                mandatory=False,
            )
        if opts.get(b'phases'):
            target_phase = phases.draft
            for head in outgoing.ancestorsof:
                target_phase = max(target_phase, repo[head].phase())
            if target_phase > phases.draft:
                part.addparam(
                    b'targetphase',
                    b'%d' % target_phase,
                    mandatory=False,
                )
    if i_repo.REPO_FEATURE_SIDE_DATA in repo.features:
        part.addparam(b'exp-sidedata', b'1')

    if opts.get(b'tagsfnodescache', True):
        addparttagsfnodescache(repo, bundler, outgoing)

    if opts.get(b'revbranchcache', True):
        addpartrevbranchcache(repo, bundler, outgoing)

    if opts.get(b'obsolescence', False):
        obsmarkers = repo.obsstore.relevantmarkers(nodes=outgoing.missing)
        buildobsmarkerspart(
            bundler,
            obsmarkers,
            mandatory=opts.get(b'obsolescence-mandatory', True),
        )

    if opts.get(b'phases', False):
        headsbyphase = phases.subsetphaseheads(repo, outgoing.missing)
        phasedata = phases.binaryencode(headsbyphase)
        bundler.newpart(b'phase-heads', data=phasedata)


def addparttagsfnodescache(repo, bundler, outgoing):
    # we include the tags fnode cache for the bundle changeset
    # (as an optional parts)
    cache = tags.hgtagsfnodescache(repo.unfiltered())
    chunks = []

    # .hgtags fnodes are only relevant for head changesets. While we could
    # transfer values for all known nodes, there will likely be little to
    # no benefit.
    #
    # We don't bother using a generator to produce output data because
    # a) we only have 40 bytes per head and even esoteric numbers of heads
    # consume little memory (1M heads is 40MB) b) we don't want to send the
    # part if we don't have entries and knowing if we have entries requires
    # cache lookups.
    for node in outgoing.ancestorsof:
        # Don't compute missing, as this may slow down serving.
        fnode = cache.getfnode(node, computemissing=False)
        if fnode:
            chunks.extend([node, fnode])

    if chunks:
        bundler.newpart(
            b'hgtagsfnodes',
            mandatory=False,
            data=b''.join(chunks),
        )


def addpartrevbranchcache(repo, bundler, outgoing):
    # we include the rev branch cache for the bundle changeset
    # (as an optional parts)
    cache = repo.revbranchcache()
    cl = repo.unfiltered().changelog
    branchesdata = collections.defaultdict(lambda: (set(), set()))
    for node in outgoing.missing:
        branch, close = cache.branchinfo(cl.rev(node))
        branchesdata[branch][close].add(node)

    def generate():
        for branch, (nodes, closed) in sorted(branchesdata.items()):
            utf8branch = encoding.fromlocal(branch)
            yield rbcstruct.pack(len(utf8branch), len(nodes), len(closed))
            yield utf8branch
            yield from sorted(nodes)
            yield from sorted(closed)

    bundler.newpart(b'cache:rev-branch-cache', data=generate(), mandatory=False)


def _formatrequirementsspec(requirements):
    requirements = [req for req in requirements if req != b"shared"]
    return urlreq.quote(b','.join(sorted(requirements)))


def _formatrequirementsparams(requirements):
    requirements = _formatrequirementsspec(requirements)
    params = b"%s%s" % (urlreq.quote(b"requirements="), requirements)
    return params


def format_remote_wanted_sidedata(repo):
    """Formats a repo's wanted sidedata categories into a bytestring for
    capabilities exchange."""
    wanted = b""
    if repo._wanted_sidedata:
        wanted = b','.join(
            pycompat.bytestr(c) for c in sorted(repo._wanted_sidedata)
        )
    return wanted


def addpartbundlestream2(
    bundler,
    repo,
    narrow_matcher: MatcherT | None = None,
    stream: bool = False,
    store_fingerprint: bytes | None = None,
    **kwargs,
):
    if not stream:
        return
    if not streamclone.allowservergeneration(repo):
        msg = _(b'stream data requested but server does not allow this feature')
        hint = _(b'the client seems buggy')
        raise error.Abort(msg, hint=hint)
    if b'stream' not in bundler.capabilities:
        msg = _(
            b'stream data requested but supported streaming clone versions '
            b'were not specified'
        )
        hint = _(b'the client seems buggy')
        raise error.Abort(msg, hint=hint)
    client_supported = set(bundler.capabilities[b'stream'])
    server_supported = set(
        bundle_caps.get_repo_caps(repo, role=b'client').get(b'stream', [])
    )
    common_supported = client_supported & server_supported
    if not common_supported:
        msg = _(b'no common supported version with the client: %s; %s')
        str_server = b','.join(sorted(server_supported))
        str_client = b','.join(sorted(client_supported))
        msg %= (str_server, str_client)
        raise error.Abort(msg)
    version = max(common_supported)

    # Stream clones don't compress well. And compression undermines a
    # goal of stream clones, which is to be fast. Communicate the desire
    # to avoid compression to consumers of the bundle.
    bundler.prefercompressed = False

    support_narrow_stream = repo.ui.configbool(
        b'experimental', b'server.stream-narrow-clones'
    )
    if narrow_matcher is not None and not support_narrow_stream:
        raise error.Abort(_(b'server does not support narrow stream clones'))

    includeobsmarkers = False
    if repo.obsstore:
        remoteversions = obsmarkersversion(bundler.capabilities)
        if not remoteversions:
            raise error.Abort(
                _(
                    b'server has obsolescence markers, but client '
                    b'cannot receive them via stream clone'
                )
            )
        elif repo.obsstore._version in remoteversions:
            includeobsmarkers = True

    if version == b"v2":
        filecount, bytecount, it = streamclone.generatev2(
            repo, narrow_matcher, includeobsmarkers
        )
        requirements = streamclone.streamed_requirements(repo)
        requirements = _formatrequirementsspec(requirements)
        part = bundler.newpart(b'stream2', data=it)
        part.addparam(b'bytecount', b'%d' % bytecount, mandatory=True)
        part.addparam(b'filecount', b'%d' % filecount, mandatory=True)
        part.addparam(b'requirements', requirements, mandatory=True)
        if store_fingerprint is not None:
            # TODO make mandatory for stream-v3
            part.addparam(
                b'store-fingerprint', store_fingerprint, mandatory=False
            )
    elif version == b"v3-exp":
        it = streamclone.generatev3(repo, narrow_matcher, includeobsmarkers)
        requirements = streamclone.streamed_requirements(repo)
        requirements = _formatrequirementsspec(requirements)
        part = bundler.newpart(b'stream3-exp', data=it)
        part.addparam(b'requirements', requirements, mandatory=True)


def buildobsmarkerspart(bundler, markers, mandatory=True):
    """add an obsmarker part to the bundler with <markers>

    No part is created if markers is empty.
    Raises ValueError if the bundler doesn't support any known obsmarker format.
    """
    if not markers:
        return None

    remoteversions = obsmarkersversion(bundler.capabilities)
    version = obsolete.commonversion(remoteversions)
    if version is None:
        raise ValueError(b'bundler does not support common obsmarker format')
    stream = obsolete.encodemarkers(markers, True, version=version)
    return bundler.newpart(b'obsmarkers', data=stream, mandatory=mandatory)


def writebundle(
    ui, cg, filename, bundletype, vfs=None, compression=None, compopts=None
):
    """Write a bundle file and return its filename.

    Existing files will not be overwritten.
    If no filename is specified, a temporary file is created.
    bz2 compression can be turned off.
    The bundle file will be deleted in case of errors.
    """

    if bundletype == b"HG20":
        bundle = bundle20(ui)
        bundle.setcompression(compression, compopts)
        part = bundle.newpart(b'changegroup', data=cg.getchunks())
        part.addparam(b'version', cg.version)
        if b'clcount' in cg.extras:
            part.addparam(
                b'nbchanges', b'%d' % cg.extras[b'clcount'], mandatory=False
            )
        chunkiter = bundle.getchunks()
    else:
        # compression argument is only for the bundle2 case
        assert compression is None
        if cg.version != b'01':
            raise error.Abort(
                _(b'old bundle types only supports v1 changegroups')
            )

        # HG20 is the case without 2 values to unpack, but is handled above.
        # pytype: disable=bad-unpacking
        header, comp = bundletypes[bundletype]
        # pytype: enable=bad-unpacking

        if comp not in util.compengines.supportedbundletypes:
            raise error.Abort(_(b'unknown stream compression type: %s') % comp)
        compengine = util.compengines.forbundletype(comp)

        def chunkiter():
            yield header
            yield from compengine.compressstream(cg.getchunks(), compopts)

        chunkiter = chunkiter()

    # parse the changegroup data, otherwise we will block
    # in case of sshrepo because we don't know the end of the stream
    return changegroup.writechunks(ui, chunkiter, filename, vfs=vfs)


def combinechangegroupresults(op):
    """logic to combine 0 or more addchangegroup results into one"""
    results = [r.get(b'return', 0) for r in op.records[b'changegroup']]
    changedheads = 0
    result = 1
    for ret in results:
        # If any changegroup result is 0, return 0
        if ret == 0:
            result = 0
            break
        if ret < -1:
            changedheads += ret + 1
        elif ret > 1:
            changedheads += ret - 1
    if changedheads > 0:
        result = 1 + changedheads
    elif changedheads < 0:
        result = -1 + changedheads
    return result


class AbortFromPart(error.Abort):
    """Sub-class of Abort that denotes an error from a bundle2 part."""


def read_remote_wanted_sidedata(remote):
    sidedata_categories = remote.capable(b'exp-wanted-sidedata')
    return read_wanted_sidedata(sidedata_categories)


def read_wanted_sidedata(formatted):
    if formatted:
        return set(formatted.split(b','))
    return set()


def widen_bundle(
    bundler, repo, oldmatcher, newmatcher, common, known, cgversion, ellipses
):
    """generates bundle2 for widening a narrow clone

    bundler is the bundle to which data should be added
    repo is the localrepository instance
    oldmatcher matches what the client already has
    newmatcher matches what the client needs (including what it already has)
    common is set of common heads between server and client
    known is a set of revs known on the client side (used in ellipses)
    cgversion is the changegroup version to send
    ellipses is boolean value telling whether to send ellipses data or not

    returns bundle2 of the data required for extending
    """
    commonnodes = set()
    cl = repo.changelog
    for r in repo.revs(b"::%ln", common):
        commonnodes.add(cl.node(r))
    if commonnodes:
        packer = changegroup.getbundler(
            cgversion,
            repo,
            oldmatcher=oldmatcher,
            matcher=newmatcher,
            fullnodes=commonnodes,
        )
        cgdata = packer.generate(
            {repo.nullid},
            list(commonnodes),
            False,
            b'narrow_widen',
            changelog=False,
        )

        part = bundler.newpart(b'changegroup', data=cgdata)
        part.addparam(b'version', cgversion)
        if scmutil.istreemanifest(repo):
            part.addparam(b'treemanifest', b'1')
    if i_repo.REPO_FEATURE_SIDE_DATA in repo.features:
        part.addparam(b'exp-sidedata', b'1')
        wanted = format_remote_wanted_sidedata(repo)
        part.addparam(b'exp-wanted-sidedata', wanted)

    return bundler



""" Mercurial phases support code

    ---

    Copyright 2011 Pierre-Yves David <pierre-yves.david@ens-lyon.org>
                   Logilab SA        <contact@logilab.fr>
                   Augie Fackler     <durin42@gmail.com>

    This software may be used and distributed according to the terms
    of the GNU General Public License version 2 or any later version.

    ---

This module implements most phase logic in mercurial.


Basic Concept
=============

A 'changeset phase' is an indicator that tells us how a changeset is
manipulated and communicated. The details of each phase is described
below, here we describe the properties they have in common.

Like bookmarks, phases are not stored in history and thus are not
permanent and leave no audit trail.

First, no changeset can be in two phases at once. Phases are ordered,
so they can be considered from lowest to highest. The default, lowest
phase is 'public' - this is the normal phase of existing changesets. A
child changeset can not be in a lower phase than its parents.

These phases share a hierarchy of traits:

            immutable shared
    public:     X        X
    draft:               X
    secret:

Local commits are draft by default.

Phase Movement and Exchange
===========================

Phase data is exchanged by pushkey on pull and push. Some servers have
a publish option set, we call such a server a "publishing server".
Pushing a draft changeset to a publishing server changes the phase to
public.

A small list of fact/rules define the exchange of phase:

* old client never changes server states
* pull never changes server states
* publish and old server changesets are seen as public by client
* any secret changeset seen in another repository is lowered to at
  least draft

Here is the final table summing up the 49 possible use cases of phase
exchange:

                           server
                  old     publish      non-publish
                 N   X    N   D   P    N   D   P
    old client
    pull
     N           -   X/X  -   X/D X/P  -   X/D X/P
     X           -   X/X  -   X/D X/P  -   X/D X/P
    push
     X           X/X X/X  X/P X/P X/P  X/D X/D X/P
    new client
    pull
     N           -   P/X  -   P/D P/P  -   D/D P/P
     D           -   P/X  -   P/D P/P  -   D/D P/P
     P           -   P/X  -   P/D P/P  -   P/D P/P
    push
     D           P/X P/X  P/P P/P P/P  D/D D/D P/P
     P           P/X P/X  P/P P/P P/P  P/P P/P P/P

Legend:

    A/B = final state on client / state on server

    * N = new/not present,
    * P = public,
    * D = draft,
    * X = not tracked (i.e., the old client or server has no internal
          way of recording the phase.)

    passive = only pushes


    A cell here can be read like this:

    "When a new client pushes a draft changeset (D) to a publishing
    server where it's not present (N), it's marked public on both
    sides (P/P)."

Note: old client behave as a publishing server with draft only content
- other people see it as public
- content is pushed as draft

"""

from __future__ import annotations

import heapq
import struct
import typing
import weakref

from typing import (
    Any,
    Callable,
    Collection,
    Iterable,
    Literal,
    overload,
)

from .i18n import _
from .node import (
    bin,
    hex,
    nullrev,
    short,
    wdirrev,
)
from .interfaces.types import (
    RepoT,
    UiT,
)
from . import (
    error,
    requirements,
    smartset,
    txnutil,
    util,
)
from .interfaces import (
    exchange as i_exc,
)

Phaseroots = dict[int, set[int]]
PhaseSets = dict[int, set[int]]

if typing.TYPE_CHECKING:
    Phasedefaults = list[Callable[[RepoT, Phaseroots], Phaseroots]]


_fphasesentry = struct.Struct(b'>i20s')

# record phase index
public: int = 0
draft: int = 1
secret: int = 2
archived = 32  # non-continuous for compatibility
internal = 96  # non-continuous for compatibility
allphases = (public, draft, secret, archived, internal)
trackedphases = (draft, secret, archived, internal)
not_public_phases = trackedphases
# record phase names
cmdphasenames = [b'public', b'draft', b'secret']  # known to `hg phase` command
phasenames = dict(enumerate(cmdphasenames))
phasenames[archived] = b'archived'
phasenames[internal] = b'internal'
# map phase name to phase number
phasenumber: dict[bytes, int] = {
    name: phase for phase, name in phasenames.items()
}
# like phasenumber, but also include maps for the numeric and binary
# phase number to the phase number
phasenumber2: dict[bytes | int, int] = phasenumber.copy()
phasenumber2.update({phase: phase for phase in phasenames})
phasenumber2.update({b'%i' % phase: phase for phase in phasenames})
# record phase property
mutablephases = (draft, secret, archived, internal)
relevant_mutable_phases = (draft, secret)  # could be obsolete or unstable
remotehiddenphases = (secret, archived, internal)
localhiddenphases = (internal, archived)

all_internal_phases = tuple(p for p in allphases if p & internal)
# We do not want any internal content to exit the repository, ever.
no_bundle_phases = all_internal_phases


def supportinternal(repo: RepoT) -> bool:
    """True if the internal phase can be used on a repository"""
    return requirements.INTERNAL_PHASE_REQUIREMENT in repo.requirements


def supportarchived(repo: RepoT) -> bool:
    """True if the archived phase can be used on a repository"""
    return requirements.ARCHIVED_PHASE_REQUIREMENT in repo.requirements


def _readroots(
    repo: RepoT,
    phasedefaults: Phasedefaults | None = None,
) -> tuple[Phaseroots, bool]:
    """Read phase roots from disk

    phasedefaults is a list of fn(repo, roots) callable, which are
    executed if the phase roots file does not exist. When phases are
    being initialized on an existing repository, this could be used to
    set selected changesets phase to something else than public.

    Return (roots, dirty) where dirty is true if roots differ from
    what is being stored.
    """
    repo = repo.unfiltered()
    dirty = False
    roots: Phaseroots = {i: set() for i in allphases}
    to_rev = repo.changelog.index.get_rev
    unknown_msg = b'removing unknown node %s from %i-phase boundary\n'
    try:
        f, pending = txnutil.trypending(repo.root, repo.svfs, b'phaseroots')
        try:
            for line in f:
                str_phase, hex_node = line.split()
                phase = int(str_phase)
                node = bin(hex_node)
                rev = to_rev(node)
                if rev is None:
                    repo.ui.debug(unknown_msg % (short(hex_node), phase))
                    dirty = True
                else:
                    roots[phase].add(rev)
        finally:
            f.close()
    except FileNotFoundError:
        if phasedefaults:
            for f in phasedefaults:
                roots = f(repo, roots)
        dirty = True
    return roots, dirty


def binaryencode(phasemapping: dict[int, list[bytes]]) -> bytes:
    """encode a 'phase -> nodes' mapping into a binary stream

    The revision lists are encoded as (phase, root) pairs.
    """
    binarydata = []
    for phase, nodes in phasemapping.items():
        for head in nodes:
            binarydata.append(_fphasesentry.pack(phase, head))
    return b''.join(binarydata)


def binarydecode(stream) -> dict[int, list[bytes]]:
    """decode a binary stream into a 'phase -> nodes' mapping

    The (phase, root) pairs are turned back into a dictionary with
    the phase as index and the aggregated roots of that phase as value."""
    headsbyphase = {i: [] for i in allphases}
    entrysize = _fphasesentry.size
    while True:
        entry = stream.read(entrysize)
        if len(entry) < entrysize:
            if entry:
                raise error.Abort(_(b'bad phase-heads stream'))
            break
        phase, node = _fphasesentry.unpack(entry)
        headsbyphase[phase].append(node)
    return headsbyphase


def _sortedrange_insert(data, idx, rev, t):
    merge_before = False
    if idx:
        r1, t1 = data[idx - 1]
        merge_before = r1[-1] + 1 == rev and t1 == t
    merge_after = False
    if idx < len(data):
        r2, t2 = data[idx]
        merge_after = r2[0] == rev + 1 and t2 == t

    if merge_before and merge_after:
        data[idx - 1] = (range(r1[0], r2[-1] + 1), t)
        data.pop(idx)
    elif merge_before:
        data[idx - 1] = (range(r1[0], rev + 1), t)
    elif merge_after:
        data[idx] = (range(rev, r2[-1] + 1), t)
    else:
        data.insert(idx, (range(rev, rev + 1), t))


def _sortedrange_split(data, idx, rev, t):
    r1, t1 = data[idx]
    if t == t1:
        return
    t = (t1[0], t[1])
    if len(r1) == 1:
        data.pop(idx)
        _sortedrange_insert(data, idx, rev, t)
    elif r1[0] == rev:
        data[idx] = (range(rev + 1, r1[-1] + 1), t1)
        _sortedrange_insert(data, idx, rev, t)
    elif r1[-1] == rev:
        data[idx] = (range(r1[0], rev), t1)
        _sortedrange_insert(data, idx + 1, rev, t)
    else:
        data[idx : idx + 1] = [
            (range(r1[0], rev), t1),
            (range(rev, rev + 1), t),
            (range(rev + 1, r1[-1] + 1), t1),
        ]


def _trackphasechange(data, rev, old, new):
    """add a phase move to the <data> list of ranges

    If data is None, nothing happens.
    """
    if data is None:
        return

    # If data is empty, create a one-revision range and done
    if not data:
        data.insert(0, (range(rev, rev + 1), (old, new)))
        return

    low = 0
    high = len(data)
    t = (old, new)
    while low < high:
        mid = (low + high) // 2
        revs = data[mid][0]
        revs_low = revs[0]
        revs_high = revs[-1]

        if rev >= revs_low and rev <= revs_high:
            _sortedrange_split(data, mid, rev, t)
            return

        if revs_low == rev + 1:
            if mid and data[mid - 1][0][-1] == rev:
                _sortedrange_split(data, mid - 1, rev, t)
            else:
                _sortedrange_insert(data, mid, rev, t)
            return

        if revs_high == rev - 1:
            if mid + 1 < len(data) and data[mid + 1][0][0] == rev:
                _sortedrange_split(data, mid + 1, rev, t)
            else:
                _sortedrange_insert(data, mid + 1, rev, t)
            return

        if revs_low > rev:
            high = mid
        else:
            low = mid + 1

    if low == len(data):
        data.append((range(rev, rev + 1), t))
        return

    r1, t1 = data[low]
    if r1[0] > rev:
        data.insert(low, (range(rev, rev + 1), t))
    else:
        data.insert(low + 1, (range(rev, rev + 1), t))


# consider incrementaly updating the phase set the update set is not bigger
# than this size
#
# Be warned, this number is picked arbitrarily, without any benchmark. It
# should blindly pickup "small update"
INCREMENTAL_PHASE_SETS_UPDATE_MAX_UPDATE = 100


class phasecache:
    if typing.TYPE_CHECKING:

        @overload
        def __init__(
            self,
            repo: Any,
            phasedefaults: Any,
            _load: Literal[False],
        ) -> None:
            pass

        @overload
        def __init__(
            self,
            repo: RepoT,
            phasedefaults: Phasedefaults | None,
            _load: bool = True,
        ) -> None:
            pass

    def __init__(
        self,
        repo,
        phasedefaults,
        _load=True,
    ):
        if _load:
            # Cheap trick to allow shallow-copy without copy module
            loaded = _readroots(repo, phasedefaults)
            self._phaseroots: Phaseroots = loaded[0]
            self.dirty: bool = loaded[1]
            self._loadedrevslen = 0
            self._phasesets: PhaseSets | None = None

    def hasnonpublicphases(self, repo: RepoT) -> bool:
        """detect if there are revisions with non-public phase"""
        # XXX deprecate the unused repo argument
        return any(
            revs for phase, revs in self._phaseroots.items() if phase != public
        )

    def nonpublicphaseroots(self, repo: RepoT) -> set[int]:
        """returns the roots of all non-public phases

        The roots are not minimized, so if the secret revisions are
        descendants of draft revisions, their roots will still be present.
        """
        repo = repo.unfiltered()
        self._ensure_phase_sets(repo)
        return set().union(
            *[
                revs
                for phase, revs in self._phaseroots.items()
                if phase != public
            ]
        )

    def get_raw_set(
        self,
        repo: RepoT,
        phase: int,
    ) -> set[int]:
        """return the set of revision in that phase

        The returned set is not filtered and might contains revision filtered
        for the passed repoview.

        The returned set might be the internal one and MUST NOT be mutated to
        avoid side effect.
        """
        if phase == public:
            raise error.ProgrammingError("cannot get_set for public phase")
        self._ensure_phase_sets(repo.unfiltered())
        revs = self._phasesets.get(phase)
        if revs is None:
            return set()
        return revs

    def getrevset(
        self,
        repo: RepoT,
        phases: Iterable[int],
        subset: Any | None = None,
    ) -> Any:
        # TODO: finish typing this
        """return a smartset for the given phases"""
        self._ensure_phase_sets(repo.unfiltered())
        phases = set(phases)
        publicphase = public in phases

        if publicphase:
            # In this case, phases keeps all the *other* phases.
            phases = set(allphases).difference(phases)
            if not phases:
                return smartset.fullreposet(repo)

        # fast path: _phasesets contains the interesting sets,
        # might only need a union and post-filtering.
        revsneedscopy = False
        if len(phases) == 1:
            [p] = phases
            revs = self._phasesets[p]
            revsneedscopy = True  # Don't modify _phasesets
        else:
            # revs has the revisions in all *other* phases.
            revs = set.union(*[self._phasesets[p] for p in phases])

        def _addwdir(wdirsubset, wdirrevs):
            if wdirrev in wdirsubset and repo[None].phase() in phases:
                if revsneedscopy:
                    wdirrevs = wdirrevs.copy()
                # The working dir would never be in the # cache, but it was in
                # the subset being filtered for its phase (or filtered out,
                # depending on publicphase), so add it to the output to be
                # included (or filtered out).
                wdirrevs.add(wdirrev)
            return wdirrevs

        if not publicphase:
            if repo.changelog.filteredrevs:
                revs = revs - repo.changelog.filteredrevs

            if subset is None:
                return smartset.baseset(revs)
            else:
                revs = _addwdir(subset, revs)
                return subset & smartset.baseset(revs)
        else:
            if subset is None:
                subset = smartset.fullreposet(repo)

            revs = _addwdir(subset, revs)

            if not revs:
                return subset
            return subset.filter(lambda r: r not in revs)

    def copy(self):
        # Shallow copy meant to ensure isolation in
        # advance/retractboundary(), nothing more.
        ph = self.__class__(None, None, _load=False)
        ph._phaseroots = self._phaseroots.copy()
        ph.dirty = self.dirty
        ph._loadedrevslen = self._loadedrevslen
        if self._phasesets is None:
            ph._phasesets = None
        else:
            ph._phasesets = self._phasesets.copy()
        return ph

    def replace(self, phcache):
        """replace all values in 'self' with content of phcache"""
        for a in (
            '_phaseroots',
            'dirty',
            '_loadedrevslen',
            '_phasesets',
        ):
            setattr(self, a, getattr(phcache, a))

    def _getphaserevsnative(self, repo):
        repo = repo.unfiltered()
        return repo.changelog.computephases(self._phaseroots)

    def _computephaserevspure(self, repo):
        repo = repo.unfiltered()
        cl = repo.changelog
        self._phasesets = {phase: set() for phase in allphases}
        lowerroots = set()
        for phase in reversed(trackedphases):
            roots = self._phaseroots[phase]
            if roots:
                ps = set(cl.descendants(roots))
                for root in roots:
                    ps.add(root)
                ps.difference_update(lowerroots)
                lowerroots.update(ps)
                self._phasesets[phase] = ps
        self._loadedrevslen = len(cl)

    def _ensure_phase_sets(self, repo: RepoT) -> None:
        """ensure phase information is loaded in the object"""
        assert repo.filtername is None
        update = -1
        cl = repo.changelog
        cl_size = len(cl)
        if self._phasesets is None:
            update = 0
        else:
            if cl_size > self._loadedrevslen:
                # check if an incremental update is worth it.
                # note we need a tradeoff here because the whole logic is not
                # stored and implemented in native code nd datastructure.
                # Otherwise the incremental update woul always be a win.
                missing = cl_size - self._loadedrevslen
                if missing <= INCREMENTAL_PHASE_SETS_UPDATE_MAX_UPDATE:
                    update = self._loadedrevslen
                else:
                    update = 0

        if update == 0:
            try:
                res = self._getphaserevsnative(repo)
                self._loadedrevslen, self._phasesets = res
            except AttributeError:
                self._computephaserevspure(repo)
            assert self._loadedrevslen == len(repo.changelog)
        elif update > 0:
            # good candidate for native code
            assert update == self._loadedrevslen
            if self.hasnonpublicphases(repo):
                start = self._loadedrevslen
                get_phase = self.phase
                rev_phases = [0] * missing
                parents = cl.parentrevs
                sets = {phase: set() for phase in self._phasesets}
                for phase, roots in self._phaseroots.items():
                    # XXX should really store the max somewhere
                    for r in roots:
                        if r >= start:
                            rev_phases[r - start] = phase
                for rev in range(start, cl_size):
                    phase = rev_phases[rev - start]
                    p1, p2 = parents(rev)
                    if p1 == nullrev:
                        p1_phase = public
                    elif p1 >= start:
                        p1_phase = rev_phases[p1 - start]
                    else:
                        p1_phase = max(phase, get_phase(repo, p1))
                    if p2 == nullrev:
                        p2_phase = public
                    elif p2 >= start:
                        p2_phase = rev_phases[p2 - start]
                    else:
                        p2_phase = max(phase, get_phase(repo, p2))
                    phase = max(phase, p1_phase, p2_phase)
                    if phase > public:
                        rev_phases[rev - start] = phase
                        sets[phase].add(rev)

                # Be careful to preserve shallow-copied values: do not update
                # phaseroots values, replace them.
                for phase, extra in sets.items():
                    if extra:
                        self._phasesets[phase] = self._phasesets[phase] | extra
            self._loadedrevslen = cl_size

    def invalidate(self):
        self._loadedrevslen = 0
        self._phasesets = None

    def phase(self, repo: RepoT, rev: int) -> int:
        # We need a repo argument here to be able to build _phasesets
        # if necessary. The repository instance is not stored in
        # phasecache to avoid reference cycles. The changelog instance
        # is not stored because it is a filecache() property and can
        # be replaced without us being notified.
        if rev == nullrev:
            return public
        if rev < nullrev:
            raise ValueError(_(b'cannot lookup negative revision'))
        # double check self._loadedrevslen to avoid an extra method call as
        # python is slow for that.
        if rev >= self._loadedrevslen:
            self._ensure_phase_sets(repo.unfiltered())
        for phase in trackedphases:
            if rev in self._phasesets[phase]:
                return phase
        return public

    def write(self, repo):
        if not self.dirty:
            return
        f = repo.svfs(b'phaseroots', b'w', atomictemp=True, checkambig=True)
        try:
            self._write(repo.unfiltered(), f)
        finally:
            f.close()

    def _write(self, repo, fp):
        assert repo.filtername is None
        to_node = repo.changelog.node
        for phase, roots in self._phaseroots.items():
            for r in sorted(roots):
                h = to_node(r)
                fp.write(b'%i %s\n' % (phase, hex(h)))
        self.dirty = False

    def _updateroots(self, repo, phase, newroots, tr, invalidate=True):
        self._phaseroots[phase] = newroots
        self.dirty = True
        if invalidate:
            self.invalidate()

        assert repo.filtername is None
        wrepo = weakref.ref(repo)

        def tr_write(fp):
            repo = wrepo()
            assert repo is not None
            self._write(repo, fp)

        tr.addfilegenerator(b'phase', (b'phaseroots',), tr_write)
        tr.hookargs[b'phases_moved'] = b'1'

    def registernew(self, repo, tr, targetphase, revs):
        repo = repo.unfiltered()
        self._retractboundary(repo, tr, targetphase, [], revs=revs)
        if tr is not None and b'phases' in tr.changes:
            phasetracking = tr.changes[b'phases']
            phase = self.phase
            for rev in sorted(revs):
                revphase = phase(repo, rev)
                _trackphasechange(phasetracking, rev, None, revphase)
        repo.invalidatevolatilesets()

    def advanceboundary(
        self, repo, tr, targetphase, nodes=None, revs=None, dryrun=None
    ):
        """Set all 'nodes' to phase 'targetphase'

        Nodes with a phase lower than 'targetphase' are not affected.

        If dryrun is True, no actions will be performed

        Returns a set of revs whose phase is changed or should be changed
        """
        if targetphase == public and not self.hasnonpublicphases(repo):
            return set()
        repo = repo.unfiltered()
        cl = repo.changelog
        torev = cl.index.rev
        # Be careful to preserve shallow-copied values: do not update
        # phaseroots values, replace them.
        new_revs = set()
        if revs is not None:
            new_revs.update(revs)
        if nodes is not None:
            new_revs.update(torev(node) for node in nodes)
        if not new_revs:  # bail out early to avoid the loadphaserevs call
            return (
                set()
            )  # note: why do people call advanceboundary with nothing?

        if tr is None:
            phasetracking = None
        else:
            phasetracking = tr.changes.get(b'phases')

        affectable_phases = sorted(
            p for p in allphases if p > targetphase and self._phaseroots[p]
        )
        # filter revision already in the right phases
        candidates = new_revs
        new_revs = set()
        self._ensure_phase_sets(repo)
        for phase in affectable_phases:
            found = candidates & self._phasesets[phase]
            new_revs |= found
            candidates -= found
            if not candidates:
                break
        if not new_revs:
            return set()

        # search for affected high phase changesets and roots
        seen = set(new_revs)
        push = heapq.heappush
        pop = heapq.heappop
        parents = cl.parentrevs
        get_phase = self.phase
        changed = {}  # set of revisions to be changed
        # set of root deleted by this path
        delroots = set()
        new_roots = {p: set() for p in affectable_phases}
        new_target_roots = set()
        # revision to walk down
        revs = [-r for r in new_revs]
        heapq.heapify(revs)
        while revs:
            current = -pop(revs)
            current_phase = get_phase(repo, current)
            changed[current] = current_phase
            p1, p2 = parents(current)
            if p1 == nullrev:
                p1_phase = public
            else:
                p1_phase = get_phase(repo, p1)
            if p2 == nullrev:
                p2_phase = public
            else:
                p2_phase = get_phase(repo, p2)
            # do we have a root ?
            if current_phase != p1_phase and current_phase != p2_phase:
                # do not record phase, because we could have "duplicated"
                # roots, were one root is shadowed by the very same roots of an
                # higher phases
                delroots.add(current)
            # schedule a walk down if needed
            if p1_phase > targetphase and p1 not in seen:
                seen.add(p1)
                push(revs, -p1)
            if p2_phase > targetphase and p2 not in seen:
                seen.add(p2)
                push(revs, -p2)
            if p1_phase < targetphase and p2_phase < targetphase:
                new_target_roots.add(current)

        # the last iteration was done with the smallest value
        min_current = current
        # do we have unwalked children that might be new roots
        if (min_current + len(changed)) < len(cl):
            for r in range(min_current, len(cl)):
                if r in changed:
                    continue
                phase = get_phase(repo, r)
                if phase <= targetphase:
                    continue
                p1, p2 = parents(r)
                if not (p1 in changed or p2 in changed):
                    continue  # not affected
                if p1 != nullrev and p1 not in changed:
                    p1_phase = get_phase(repo, p1)
                    if p1_phase == phase:
                        continue  # not a root
                if p2 != nullrev and p2 not in changed:
                    p2_phase = get_phase(repo, p2)
                    if p2_phase == phase:
                        continue  # not a root
                new_roots[phase].add(r)

        # apply the changes
        if not dryrun:
            for r, p in changed.items():
                _trackphasechange(phasetracking, r, p, targetphase)
            if targetphase > public:
                self._phasesets[targetphase].update(changed)
            for phase in affectable_phases:
                roots = self._phaseroots[phase]
                removed = roots & delroots
                if removed or new_roots[phase]:
                    self._phasesets[phase].difference_update(changed)
                    # Be careful to preserve shallow-copied values: do not
                    # update phaseroots values, replace them.
                    final_roots = roots - delroots | new_roots[phase]
                    self._updateroots(
                        repo, phase, final_roots, tr, invalidate=False
                    )
            if new_target_roots:
                # Thanks for previous filtering, we can't replace existing
                # roots
                new_target_roots |= self._phaseroots[targetphase]
                self._updateroots(
                    repo, targetphase, new_target_roots, tr, invalidate=False
                )
            repo.invalidatevolatilesets()
        return changed

    def retractboundary(self, repo, tr, targetphase, nodes):
        if tr is None:
            phasetracking = None
        else:
            phasetracking = tr.changes.get(b'phases')
        repo = repo.unfiltered()
        retracted = self._retractboundary(repo, tr, targetphase, nodes)
        if retracted and phasetracking is not None:
            for r, old_phase in sorted(retracted.items()):
                _trackphasechange(phasetracking, r, old_phase, targetphase)
        repo.invalidatevolatilesets()

    def _retractboundary(self, repo, tr, targetphase, nodes=None, revs=None):
        if targetphase == public:
            return {}
        if (
            targetphase == internal
            and not supportinternal(repo)
            or targetphase == archived
            and not supportarchived(repo)
        ):
            name = phasenames[targetphase]
            msg = b'this repository does not support the %s phase' % name
            raise error.ProgrammingError(msg)
        assert repo.filtername is None
        cl = repo.changelog
        torev = cl.index.rev
        new_revs = set()
        if revs is not None:
            new_revs.update(revs)
        if nodes is not None:
            new_revs.update(torev(node) for node in nodes)
        if not new_revs:  # bail out early to avoid the loadphaserevs call
            return {}  # note: why do people call retractboundary with nothing ?

        if nullrev in new_revs:
            raise error.Abort(_(b'cannot change null revision phase'))

        # Filter revision that are already in the right phase
        self._ensure_phase_sets(repo)
        for phase, revs in self._phasesets.items():
            if phase >= targetphase:
                new_revs -= revs
        if not new_revs:  # all revisions already in the right phases
            return {}

        # Compute change in phase roots by walking the graph
        #
        # note: If we had a cheap parent → children mapping we could do
        # something even cheaper/more-bounded
        #
        # The idea would be to walk from item in new_revs stopping at
        # descendant with phases >= target_phase.
        #
        # 1) This detect new_revs that are not new_roots (either already >=
        #    target_phase or reachable though another new_revs
        # 2) This detect replaced current_roots as we reach them
        # 3) This can avoid walking to the tip if we retract over a small
        #    branch.
        #
        # So instead, we do a variation of this, we walk from the smaller new
        # revision to the tip to avoid missing any potential children.
        #
        # The following code would be a good candidate for native code… if only
        # we could knew the phase of a changeset efficiently in native code.
        parents = cl.parentrevs
        phase = self.phase
        new_roots = set()  # roots added by this phases
        changed_revs = {}  # revision affected by this call
        replaced_roots = set()  # older roots replaced by this call
        currentroots = self._phaseroots[targetphase]
        start = min(new_revs)
        end = len(cl)
        rev_phases = [None] * (end - start)

        this_phase_set = self._phasesets[targetphase]
        for r in range(start, end):
            # gather information about the current_rev
            r_phase = phase(repo, r)
            p_phase = None  # phase inherited from parents
            p1, p2 = parents(r)
            if p1 >= start:
                p1_phase = rev_phases[p1 - start]
                if p1_phase is not None:
                    p_phase = p1_phase
            if p2 >= start:
                p2_phase = rev_phases[p2 - start]
                if p2_phase is not None:
                    if p_phase is not None:
                        p_phase = max(p_phase, p2_phase)
                    else:
                        p_phase = p2_phase

            # assess the situation
            if r in new_revs and r_phase < targetphase:
                if p_phase is None or p_phase < targetphase:
                    new_roots.add(r)
                rev_phases[r - start] = targetphase
                changed_revs[r] = r_phase
                this_phase_set.add(r)
            elif p_phase is None:
                rev_phases[r - start] = r_phase
            else:
                if p_phase > r_phase:
                    rev_phases[r - start] = p_phase
                else:
                    rev_phases[r - start] = r_phase
                if p_phase == targetphase:
                    if p_phase > r_phase:
                        changed_revs[r] = r_phase
                        this_phase_set.add(r)
                    elif r in currentroots:
                        replaced_roots.add(r)
            sets = self._phasesets
            if targetphase > draft:
                for r, old in changed_revs.items():
                    if old > public:
                        sets[old].discard(r)

        if new_roots:
            assert changed_revs

            final_roots = new_roots | currentroots - replaced_roots
            self._updateroots(
                repo,
                targetphase,
                final_roots,
                tr,
                invalidate=False,
            )
            if targetphase > 1:
                retracted = set(changed_revs)
                for lower_phase in range(1, targetphase):
                    lower_roots = self._phaseroots.get(lower_phase)
                    if lower_roots is None:
                        continue
                    if lower_roots & retracted:
                        simpler_roots = lower_roots - retracted
                        self._updateroots(
                            repo,
                            lower_phase,
                            simpler_roots,
                            tr,
                            invalidate=False,
                        )
            return changed_revs
        else:
            assert not changed_revs
            assert not replaced_roots
            return {}

    def register_strip(
        self,
        repo,
        tr,
        strip_rev: int,
    ):
        """announce a strip to the phase cache

        Any roots higher than the stripped revision should be dropped.
        """
        for targetphase, roots in list(self._phaseroots.items()):
            filtered = {r for r in roots if r >= strip_rev}
            if filtered:
                self._updateroots(repo, targetphase, roots - filtered, tr)
        self.invalidate()


def advanceboundary(repo, tr, targetphase, nodes, revs=None, dryrun=None):
    """Add nodes to a phase changing other nodes phases if necessary.

    This function move boundary *forward* this means that all nodes
    are set in the target phase or kept in a *lower* phase.

    Simplify boundary to contains phase roots only.

    If dryrun is True, no actions will be performed

    Returns a set of revs whose phase is changed or should be changed
    """
    if revs is None:
        revs = []
    phcache = repo._phasecache.copy()
    changes = phcache.advanceboundary(
        repo, tr, targetphase, nodes, revs=revs, dryrun=dryrun
    )
    if not dryrun:
        repo._phasecache.replace(phcache)
    return changes


def retractboundary(repo, tr, targetphase, nodes):
    """Set nodes back to a phase changing other nodes phases if
    necessary.

    This function move boundary *backward* this means that all nodes
    are set in the target phase or kept in a *higher* phase.

    Simplify boundary to contains phase roots only."""
    phcache = repo._phasecache.copy()
    phcache.retractboundary(repo, tr, targetphase, nodes)
    repo._phasecache.replace(phcache)


def registernew(repo, tr, targetphase, revs):
    """register a new revision and its phase

    Code adding revisions to the repository should use this function to
    set new changeset in their target phase (or higher).
    """
    phcache = repo._phasecache.copy()
    phcache.registernew(repo, tr, targetphase, revs)
    repo._phasecache.replace(phcache)


def listphases(repo: RepoT) -> dict[bytes, bytes]:
    """List phases root for serialization over pushkey"""
    # Use ordered dictionary so behavior is deterministic.
    keys = util.sortdict()
    value = b'%i' % draft
    cl = repo.unfiltered().changelog
    to_node = cl.node
    for root in repo._phasecache._phaseroots[draft]:
        if repo._phasecache.phase(repo, root) <= draft:
            keys[hex(to_node(root))] = value

    if repo.publishing():
        # Add an extra data to let remote know we are a publishing
        # repo. Publishing repo can't just pretend they are old repo.
        # When pushing to a publishing repo, the client still need to
        # push phase boundary
        #
        # Push do not only push changeset. It also push phase data.
        # New phase data may apply to common changeset which won't be
        # push (as they are common). Here is a very simple example:
        #
        # 1) repo A push changeset X as draft to repo B
        # 2) repo B make changeset X public
        # 3) repo B push to repo A. X is not pushed but the data that
        #    X as now public should
        #
        # The server can't handle it on it's own as it has no idea of
        # client phase data.
        keys[b'publishing'] = b'True'
    return keys


def pushphase(
    repo: RepoT,
    nhex: bytes,
    oldphasestr: bytes,
    newphasestr: bytes,
) -> bool:
    """List phases root for serialization over pushkey"""
    repo = repo.unfiltered()
    with repo.lock():
        currentphase = repo[nhex].phase()
        newphase = abs(int(newphasestr))  # let's avoid negative index surprise
        oldphase = abs(int(oldphasestr))  # let's avoid negative index surprise
        if currentphase == oldphase and newphase < oldphase:
            with repo.transaction(b'pushkey-phase') as tr:
                advanceboundary(repo, tr, newphase, [bin(nhex)])
            return True
        elif currentphase == newphase:
            # raced, but got correct result
            return True
        else:
            return False


def subsetphaseheads(repo, subset):
    """Finds the phase heads for a subset of a history

    Returns a list indexed by phase number where each item is a list of phase
    head nodes.
    """
    cl = repo.changelog

    headsbyphase = {i: [] for i in allphases}
    for phase in allphases:
        revset = b"heads(%%ln & _phase(%d))" % phase
        headsbyphase[phase] = [cl.node(r) for r in repo.revs(revset, subset)]
    return headsbyphase


def updatephases(repo, trgetter, headsbyphase):
    """Updates the repo with the given phase heads"""
    # Now advance phase boundaries of all phases
    #
    # run the update (and fetch transaction) only if there are actually things
    # to update. This avoid creating empty transaction during no-op operation.

    for phase in allphases:
        revset = b'%ln - _phase(%s)'
        heads = [c.node() for c in repo.set(revset, headsbyphase[phase], phase)]
        if heads:
            advanceboundary(repo, trgetter(), phase, heads)


def analyze_remote_phases(
    repo,
    subset: Collection[int],
    roots: dict[bytes, bytes],
) -> tuple[Collection[int], Collection[int]]:
    """Compute phases heads and root in a subset of node from root dict

    * subset is heads of the subset
    * roots is {<nodeid> => phase} mapping. key and value are string.

    Accept unknown element input
    """
    repo = repo.unfiltered()
    # build list from dictionary
    draft_roots = []
    to_rev = repo.changelog.index.get_rev
    for nhex, phase in roots.items():
        if nhex == b'publishing':  # ignore data related to publish option
            continue
        node = bin(nhex)
        phase = int(phase)
        if phase == public:
            if node != repo.nullid:
                msg = _(b'ignoring inconsistent public root from remote: %s\n')
                repo.ui.warn(msg % nhex)
        elif phase == draft:
            rev = to_rev(node)
            if rev is not None:  # to filter unknown nodes
                draft_roots.append(rev)
        else:
            msg = _(b'ignoring unexpected root from remote: %i %s\n')
            repo.ui.warn(msg % (phase, nhex))
    # compute heads
    public_heads = new_heads(repo, subset, draft_roots)
    return public_heads, draft_roots


class RemotePhasesSummary(i_exc.IRemotePhasesSummary):
    """summarize phase information on the remote side

    :publishing: True is the remote is publishing
    :public_heads: list of remote public phase heads (revs)
    :draft_heads: list of remote draft phase heads (revs)
    :draft_roots: list of remote draft phase root (revs)
    """

    def __init__(
        self,
        repo,
        remote_subset: Collection[int],
        remote_roots: dict[bytes, bytes],
    ):
        unfi = repo.unfiltered()
        self._allremoteroots: dict[bytes, bytes] = remote_roots

        self.publishing: bool = bool(remote_roots.get(b'publishing', False))

        heads, roots = analyze_remote_phases(repo, remote_subset, remote_roots)
        self.public_heads: Collection[int] = heads
        self.draft_roots: Collection[int] = roots
        # Get the list of all "heads" revs draft on remote
        dheads = unfi.revs(b'heads(%ld::%ld)', roots, remote_subset)
        self.draft_heads: Collection[int] = dheads


def new_heads(
    repo,
    heads: Collection[int],
    roots: Collection[int],
) -> Collection[int]:
    """compute new head of a subset minus another

    * `heads`: define the first subset
    * `roots`: define the second we subtract from the first"""

    if not roots:
        return heads
    if not heads or heads == [nullrev]:
        return []
    # The logic operated on revisions, convert arguments early for convenience
    # PERF-XXX: maybe heads could directly comes as a set without impacting
    # other user of that value
    new_heads = set(heads)
    new_heads.discard(nullrev)
    # compute the area we need to remove
    affected_zone = repo.revs(b"(%ld::%ld)", roots, new_heads)
    # heads in the area are no longer heads
    new_heads.difference_update(affected_zone)
    # revisions in the area have children outside of it,
    # They might be new heads
    candidates = repo.revs(
        b"parents(%ld + (%ld and merge())) and not null", roots, affected_zone
    )
    candidates -= affected_zone
    if candidates:
        # remove candidate that are ancestors of other heads
        new_heads.update(candidates)
        prunestart = repo.revs(b"parents(%ld) and not null", new_heads)
        pruned = repo.changelog.reachableroots(
            candidates.min(),
            list(candidates),
            list(prunestart),
        )
        new_heads.difference_update(pruned)

    # PERF-XXX: do we actually need a sorted list here? Could we simply return
    # a set?
    return sorted(new_heads)


def newcommitphase(ui: UiT) -> int:
    """helper to get the target phase of new commit

    Handle all possible values for the phases.new-commit options.

    """
    v = ui.config(b'phases', b'new-commit')
    try:
        return phasenumber2[v]
    except KeyError:
        raise error.ConfigError(
            _(b"phases.new-commit: not a valid phase name ('%s')") % v
        )


def hassecret(repo: RepoT) -> bool:
    """utility function that check if a repo have any secret changeset."""
    return bool(repo._phasecache._phaseroots[secret])


def preparehookargs(
    node: bytes,
    old: int | None,
    new: int | None,
) -> dict[bytes, bytes]:
    if old is None:
        old = b''
    else:
        old = phasenames[old]
    return {b'node': node, b'oldphase': old, b'phase': phasenames[new]}
